# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

# AA TaxSystem
from taxsystem.api.helpers import generate_button, get_info_button
from taxsystem.models.tax import Payments


def _approve_payment(actions: list, corporation_id: int, payment: Payments, request):
    amount = intcomma(payment.amount, use_l10n=True)
    actions.append(
        generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=payment,
            settings={
                "title": _("Approve Payment"),
                "icon": "fas fa-check",
                "color": "success",
                "text": _("Approve Payment")
                + f" {amount} ISK "
                + _("from")
                + f" {payment.account.user.username}",
                "modal": "payments-approve",
                "action": reverse(
                    viewname="taxsystem:approve_payment",
                    kwargs={
                        "corporation_id": corporation_id,
                        "payment_pk": payment.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )
    )


def _reject_payment(actions: list, corporation_id: int, payment: Payments, request):
    amount = intcomma(payment.amount, use_l10n=True)
    actions.append(
        generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=payment,
            settings={
                "title": _("Reject Payment"),
                "icon": "fas fa-times",
                "color": "danger",
                "text": _("Reject Payment")
                + f" {amount} ISK "
                + _("from")
                + f" {payment.account.user.username}",
                "modal": "payments-reject",
                "action": reverse(
                    viewname="taxsystem:reject_payment",
                    kwargs={
                        "corporation_id": corporation_id,
                        "payment_pk": payment.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )
    )


def _undo_payment(actions: list, corporation_id: int, payment: Payments, request):
    amount = intcomma(payment.amount, use_l10n=True)
    actions.append(
        generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=payment,
            settings={
                "title": _("Undo Payment"),
                "icon": "fas fa-undo",
                "color": "warning",
                "text": _("Undo Payment")
                + f" {amount} ISK "
                + _("from")
                + f" {payment.account.user.username}",
                "modal": "payments-undo",
                "action": reverse(
                    viewname="taxsystem:undo_payment",
                    kwargs={
                        "corporation_id": corporation_id,
                        "payment_pk": payment.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )
    )


def _delete_payment(actions: list, corporation_id: int, payment: Payments, request):
    amount = intcomma(payment.amount, use_l10n=True)
    actions.append(
        generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=payment,
            settings={
                "title": _("Delete Payment"),
                "icon": "fas fa-trash",
                "color": "danger",
                "text": _("Delete Payment")
                + f" {amount} ISK "
                + _("from")
                + f" {payment.account.user.username}",
                "modal": "payments-delete",
                "action": reverse(
                    viewname="taxsystem:delete_payment",
                    kwargs={
                        "corporation_id": corporation_id,
                        "payment_pk": payment.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )
    )


def payments_actions(corporation_id, payment: Payments, perms, request):
    actions = []
    if perms:
        if payment.is_pending or payment.is_needs_approval:
            _approve_payment(actions, corporation_id, payment, request)
            _reject_payment(actions, corporation_id, payment, request)
        elif payment.is_approved or payment.is_rejected:
            _undo_payment(actions, corporation_id, payment, request)
    if payment.account.user == request.user or perms:
        # Only allow deleting payments that are not ESI recorded
        if payment.entry_id == 0:
            _delete_payment(actions, corporation_id, payment, request)

        actions.append(get_info_button(corporation_id, payment, request))

    actions_html = format_html("".join(actions))
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)



"""Custom exceptions."""

# Alliance Auth
from esi.errors import TokenError


class TokenDoesNotExist(TokenError):
    """A token with a specific scope does not exist for a user."""


class DatabaseError(Exception):
    pass


class NotModifiedError(Exception):
    pass


class HTTPGatewayTimeoutError(Exception):
    pass



# AA Tax System
# AA TaxSystem
from taxsystem.models.filters import JournalFilter, JournalFilterSet
from taxsystem.models.tax import OwnerAudit


def create_filterset(owner: OwnerAudit, **kwargs) -> JournalFilterSet:
    """Create a FilterSet for a Corporation"""
    params = {
        "owner": owner,
    }
    params.update(kwargs)
    journal_filter_set = JournalFilterSet(**params)
    journal_filter_set.save()
    return journal_filter_set


def create_filter(filter_set: JournalFilterSet, **kwargs) -> JournalFilter:
    """Create a Filter for a Corporation"""
    params = {
        "filter_set": filter_set,
    }
    params.update(kwargs)
    journal_filter = JournalFilter(**params)
    journal_filter.save()
    return journal_filter



"""Logs Model"""

# Django
from django.db import models
from django.utils import timezone
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import User

# AA TaxSystem
from taxsystem.managers.logs_manager import LogsManager
from taxsystem.models.tax import OwnerAudit, Payments


class PaymentHistory(models.Model):
    """PaymentHistory Model for app"""

    class SystemText(models.TextChoices):
        DEFAULT = "", ""
        ADDED = "Payment added to system", _("Payment added to system")
        AUTOMATIC = "Automated approved Payment", _("Automated approved Payment")
        REVISER = "Payment must be approved by an reviser", _(
            "Payment must be approved by an reviser"
        )

    class Actions(models.TextChoices):
        DEFAULT = "", ""
        STATUS_CHANGE = "Status Changed", _("Status Changed")
        PAYMENT_ADDED = "Payment Added", _("Payment Added")
        REVISER_COMMENT = "Reviser Comment", _("Reviser Comment")

    class Meta:
        default_permissions = ()

    objects = LogsManager()

    payment = models.ForeignKey(
        Payments,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Payment"),
        help_text=_("Payment that the action was performed on"),
    )

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("User"),
        help_text=_("User that performed the action"),
    )

    date = models.DateTimeField(
        default=timezone.now,
        verbose_name=_("Date"),
        help_text=_("Date of the action"),
    )

    action = models.CharField(
        max_length=20,
        choices=Actions.choices,
        default=Actions.DEFAULT,
        verbose_name=_("Action"),
        help_text=_("Action performed"),
    )

    comment = models.CharField(
        max_length=255,
        choices=SystemText.choices,
        default=SystemText.DEFAULT,
        verbose_name=_("Comment"),
        help_text=_("Comment of the action"),
    )

    new_status = models.CharField(
        max_length=16,
        choices=Payments.RequestStatus.choices,
        verbose_name=_("New Status"),
        help_text=_("New Status of the action"),
    )

    def __str__(self):
        return f"{self.date}: {self.user} - {self.action} - {self.comment}"


class AdminLogs(models.Model):
    """Logs Model for app"""

    class Actions(models.TextChoices):
        DEFAULT = "", ""
        ADD = "Added", _("Added")
        CHANGE = "Changed", _("Changed")
        DELETE = "Deleted", _("Deleted")

    class Meta:
        default_permissions = ()

    objects = LogsManager()

    owner = models.ForeignKey(
        OwnerAudit,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("Payment"),
        help_text=_("Payment that the action was performed on"),
    )

    user = models.ForeignKey(
        User,
        on_delete=models.CASCADE,
        related_name="+",
        verbose_name=_("User"),
        help_text=_("User that performed the action"),
    )

    date = models.DateTimeField(
        default=timezone.now,
        verbose_name=_("Date"),
        help_text=_("Date of the action"),
    )

    action = models.CharField(
        max_length=20,
        choices=Actions.choices,
        default=Actions.DEFAULT,
        verbose_name=_("Action"),
        help_text=_("Action performed"),
    )

    comment = models.TextField(
        blank=True,
        default="",
        verbose_name=_("Comment"),
    )

    def __str__(self):
        return f"{self.date}: {self.user} - {self.action} - {self.comment}"



"""App Configuration"""

# Django
from django.apps import AppConfig

# AA TaxSystem
from taxsystem import __version__


class TaxSystemConfig(AppConfig):
    """App Config"""

    default_auto_field = "django.db.models.AutoField"
    name = "taxsystem"
    label = "taxsystem"
    verbose_name = f"Tax System v{__version__}"



# Standard Library
import logging

# Third Party
from ninja import NinjaAPI
from ninja.security import django_auth

# Django
from django.conf import settings

# AA TaxSystem
from taxsystem.api import taxsystem

logger = logging.getLogger(__name__)

api = NinjaAPI(
    title="Geuthur API",
    version="0.1.0",
    urls_namespace="taxsystem:api",
    auth=django_auth,
    csrf=True,
    openapi_url=settings.DEBUG and "/openapi.json" or "",
)

# Add the taxsystem endpoints
taxsystem.setup(api)



from .admin import AdminApiEndpoints
from .character import CharacterApiEndpoints
from .corporation import CorporationApiEndpoints


def setup(api):
    AdminApiEndpoints(api)
    CorporationApiEndpoints(api)
    CharacterApiEndpoints(api)



"""TestView class."""

# Standard Library
import json
from http import HTTPStatus

# Django
from django.test import RequestFactory, TestCase
from django.urls import reverse

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# AA TaxSystem
from taxsystem import views
from taxsystem.models.tax import Payments
from taxsystem.tests.testdata.generate_owneraudit import (
    create_owneraudit_from_user,
)
from taxsystem.tests.testdata.generate_payments import (
    create_payment,
    create_payment_system,
)
from taxsystem.tests.testdata.load_allianceauth import load_allianceauth
from taxsystem.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "taxsystem.views"


class TestAddPayment(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.factory = RequestFactory()
        cls.user, cls.character_ownership = create_user_from_evecharacter(
            character_id=1001,
            permissions=[
                "taxsystem.basic_access",
                "taxsystem.manage_own_corp",
            ],
        )
        cls.audit = create_owneraudit_from_user(cls.user)
        cls.no_audit_user, _ = create_user_from_evecharacter(
            character_id=1002,
            permissions=[
                "taxsystem.basic_access",
                "taxsystem.manage_own_corp",
            ],
        )
        cls.no_permission_user, _ = create_user_from_evecharacter(
            character_id=1003,
            permissions=[
                "taxsystem.basic_access",
            ],
        )
        cls.payment_system = create_payment_system(
            name=cls.user.username,
            owner=cls.audit,
            user=cls.user,
        )

    def test_add_payment(self):
        """Test adding a payment."""
        # given
        corporation_id = self.audit.corporation.corporation_id
        form_data = {
            "corporation_id": corporation_id,
            "add_reason": "This is a test addition",
            "amount": 1000,
        }
        request = self.factory.post(
            reverse(
                "taxsystem:add_payment", args=[corporation_id, self.payment_system.pk]
            ),
            data=form_data,
        )
        request.user = self.user
        # when
        response = views.add_payment(
            request,
            corporation_id=corporation_id,
            payment_system_pk=self.payment_system.pk,
        )
        payment = Payments.objects.get(
            account=self.payment_system, amount=1000, reason="This is a test addition"
        )
        response_data = json.loads(response.content)
        # then
        self.assertEqual(response.status_code, HTTPStatus.OK)
        self.assertTrue(response_data["success"])
        self.assertEqual(
            response_data["message"],
            f"Payment ID: {payment.pk} - Amount: 1,000 - Name: Gneuten added",
        )

    def test_no_permission(self):
        """Test try adding a payment without permission."""
        # given
        corporation_id = self.audit.corporation.corporation_id
        form_data = {
            "corporation_id": corporation_id,
            "add_reason": "This is a test addition",
            "amount": 1000,
        }
        request = self.factory.post(
            reverse(
                "taxsystem:add_payment", args=[corporation_id, self.payment_system.pk]
            ),
            data=form_data,
        )
        request.user = self.no_audit_user
        # when
        response = views.add_payment(
            request,
            corporation_id=corporation_id,
            payment_system_pk=self.payment_system.pk,
        )
        response_data = json.loads(response.content)
        # then
        self.assertEqual(response.status_code, HTTPStatus.FORBIDDEN)
        self.assertFalse(response_data["success"])
        self.assertEqual(response_data["message"], "Permission Denied")

    def test_no_manage_permission(self):
        """Test reject payment without managing permission."""
        # given
        corporation_id = self.audit.corporation.corporation_id
        form_data = {
            "corporation_id": corporation_id,
            "add_reason": "This is a test addition",
            "amount": 1000,
        }
        # when
        request = self.factory.post(
            reverse(
                "taxsystem:add_payment", args=[corporation_id, self.payment_system.pk]
            ),
            data=form_data,
        )
        request.user = self.no_permission_user
        # then
        response = views.add_payment(
            request,
            corporation_id=corporation_id,
            payment_system_pk=self.payment_system.pk,
        )

        self.assertEqual(response.status_code, HTTPStatus.FOUND)



# Django
from django.contrib.auth.models import Permission
from django.test import TestCase
from django.utils import timezone

# Alliance Auth
from allianceauth.tests.auth_utils import AuthUtils

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter

# AA TaxSystem
from taxsystem.models.tax import PaymentSystem
from taxsystem.tests.testdata.generate_owneraudit import (
    create_owneraudit_from_user,
)
from taxsystem.tests.testdata.generate_payments import (
    create_payment_system,
)
from taxsystem.tests.testdata.load_allianceauth import load_allianceauth

MODULE_PATH = "taxsystem.models.tax"


class TestPaymentSystemModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["taxsystem.basic_access"]
        )
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter(
            1002, permissions=["taxsystem.basic_access"]
        )
        cls.audit = create_owneraudit_from_user(cls.user)
        cls.audit2 = create_owneraudit_from_user(cls.user2)

        cls.payment_system = create_payment_system(
            name=cls.user.username,
            owner=cls.audit,
            user=cls.user,
        )

    def test_str(self):
        expected_str = PaymentSystem.objects.get(owner=self.audit)
        self.assertEqual(self.payment_system, expected_str)

    def test_is_active(self):
        """Test if the payment system is active."""
        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertTrue(payment_system.is_active)

    def test_is_inactive(self):
        """Test if the payment system is inactive."""
        self.payment_system.status = PaymentSystem.Status.INACTIVE
        self.payment_system.save()

        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertFalse(payment_system.is_active)

    def test_is_deactivated(self):
        """Test if the payment system is deactivated."""
        self.payment_system.status = PaymentSystem.Status.DEACTIVATED
        self.payment_system.save()
        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertFalse(payment_system.is_active)

    def test_is_missing(self):
        """Test if the payment system is missing."""
        self.payment_system.status = PaymentSystem.Status.MISSING
        self.payment_system.save()

        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertFalse(payment_system.is_active)

    def test_has_paid(self):
        """Test if the payment system has paid."""
        self.payment_system.deposit = 1000
        self.payment_system.save()

        self.payment_system.date = timezone.now()

        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertTrue(payment_system.has_paid)

    def test_has_paid_icon(self):
        """Test the icon representation of has_paid."""
        self.payment_system.deposit = 1000
        self.payment_system.save()

        self.payment_system.date = timezone.now()

        payment_system = PaymentSystem.objects.get(owner=self.audit)

        self.assertIn(
            "fas fa-check",
            payment_system.has_paid_icon()["display"],
        )
        self.assertIn(
            "badge",
            payment_system.has_paid_icon(badge=True, text=True)["display"],
        )
        self.assertIn(
            "Paid",
            payment_system.has_paid_icon(badge=True, text=True)["display"],
        )

    def test_status_html(self):
        """Test the HTML representation of the payment system status."""
        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertIn(
            "bg-success",
            payment_system.Status(payment_system.status).html(),
        )

        self.assertIn(
            "Active",
            payment_system.Status(payment_system.status).html(text=True),
        )

    def test_status_color(self):
        """Test the color representation of the payment system status."""
        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertEqual(
            payment_system.Status(payment_system.status).color(), "success"
        )

    def test_status_icon(self):
        """Test the icon representation of the payment system status."""
        payment_system = PaymentSystem.objects.get(owner=self.audit)

        self.assertEqual(
            payment_system.Status(payment_system.status).icons(),
            "<i class='fas fa-check'></i>",
        )

    def test_deposit_html(self):
        """Test the HTML representation of the deposit."""
        payment_system = PaymentSystem.objects.get(owner=self.audit)
        self.assertIn("0 ISK", payment_system.deposit_html)

        payment_system.deposit = 1500000
        payment_system.save()
        self.assertIn(
            "<span class='text-success'>1,500,000 ISK</span>",
            payment_system.deposit_html,
        )

        payment_system = PaymentSystem.objects.get(owner=self.audit)
        payment_system.deposit = -500000
        payment_system.save()
        self.assertIn(
            "<span class='text-danger'>-500,000 ISK</span>", payment_system.deposit_html
        )



# Django
from django.utils.html import format_html
from django.utils.translation import gettext as _

# AA TaxSystem
from taxsystem.api.helpers import get_info_button
from taxsystem.models.tax import Payments


def _own_payments_actions(corporation_id, payment: Payments, request):
    button = get_info_button(corporation_id, payment, request)
    return format_html('<div class="d-flex justify-content-end">{}</div>', button)



from .filters import JournalFilter, JournalFilterSet
from .general import General
from .logs import AdminLogs, PaymentHistory
from .tax import (
    Members,
    OwnerAudit,
    OwnerUpdateStatus,
    Payments,
)
from .wallet import CorporationWalletDivision, CorporationWalletJournalEntry



"""Forms for the taxsystem app."""

# Django
from django import forms
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.models.filters import JournalFilter, JournalFilterSet


def get_mandatory_form_label_text(text: str) -> str:
    """Label text for mandatory form fields"""

    required_marker = "<span class='form-required-marker'>*</span>"

    return mark_safe(
        f"<span class='form-field-required'>{text} {required_marker}</span>"
    )


class PaymentRejectForm(forms.Form):
    """Form for payment rejecting."""

    reject_reason = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Reason for rejecting")),
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class PaymentDeleteForm(forms.Form):
    """Form for payment deleting."""

    delete_reason = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Reason for deleting payment")),
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class PaymentAddForm(forms.Form):
    """Form for payment adding."""

    amount = forms.IntegerField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Amount")),
        widget=forms.NumberInput(attrs={"min": "0"}),
    )

    add_reason = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Reason for adding payment")),
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class PaymentAcceptForm(forms.Form):
    """Form for payment accepting."""

    accept_info = forms.CharField(
        required=False,
        label=_("Comment") + " (optional)",
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class PaymentUndoForm(forms.Form):
    """Form for payment undoing."""

    undo_reason = forms.CharField(
        required=True,
        label=get_mandatory_form_label_text(text=_("Reason for undoing")),
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class MemberDeleteForm(forms.Form):
    """Form for member deleting."""

    delete_reason = forms.CharField(
        required=False,
        label=_("Comment") + " (optional)",
        widget=forms.Textarea(attrs={"rows": 5}),
    )


class FilterDeleteForm(forms.Form):
    """Form for deleting."""

    delete_reason = forms.CharField(
        required=False,
        label=_("Comment") + " (optional)",
        widget=forms.Textarea(attrs={"rows": 5}),
    )

    filter = forms.HiddenInput()


class TaxSwitchUserForm(forms.Form):
    """Form for switching user."""

    user = forms.HiddenInput()


class AddJournalFilterForm(forms.Form):
    filter_set = forms.ModelChoiceField(
        queryset=None,
        label=_("Filter Set"),
        required=True,
    )
    filter_type = forms.ChoiceField(
        choices=JournalFilter.FilterType.choices,
        label=_("Filter Type"),
        required=True,
    )
    value = forms.CharField(
        label=_("Filter Value"),
        required=True,
        widget=forms.TextInput(attrs={"placeholder": _("Enter filter value")}),
    )

    def __init__(self, *args, queryset=None, **kwargs):
        super().__init__(*args, **kwargs)
        if queryset is not None:
            self.fields["filter_set"].queryset = queryset


class CreateFilterSetForm(forms.Form):
    name = forms.CharField(
        label=_("Filter Set Name"),
        required=True,
        widget=forms.TextInput(attrs={"placeholder": _("Enter filter set name")}),
    )
    description = forms.CharField(
        label=_("Filter Set Description"),
        required=False,
        widget=forms.Textarea(
            attrs={"placeholder": _("Enter filter set description"), "rows": 3}
        ),
    )


class EditFilterSetForm(forms.ModelForm):
    class Meta:
        model = JournalFilterSet
        fields = ["name", "description"]



"""This module provides lazy loading of some common functions and objects that are not needed for every request."""

# Django
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.eveonline.evelinks.eveimageserver import (
    character_portrait_url,
    corporation_logo_url,
    type_render_url,
)


def get_character_portrait_url(
    character_id: int, size: int = 32, character_name: str = None, as_html: bool = False
) -> str:
    """Get the character portrait for a character ID."""

    render_url = character_portrait_url(character_id=character_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="character-portrait rounded-circle" src="{}" alt="{}">',
            render_url,
            character_name,
        )
        return render_html
    return render_url


def get_corporation_logo_url(
    corporation_id: int,
    size: int = 32,
    corporation_name: str = None,
    as_html: bool = False,
) -> str:
    """Get the corporation logo for a corporation ID."""

    render_url = corporation_logo_url(corporation_id=corporation_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="corporation-logo rounded-circle" src="{}" alt="{}">',
            render_url,
            corporation_name,
        )
        return render_html
    return render_url


def get_type_render_url(
    type_id: int, size: int = 32, type_name: str = None, as_html: bool = False
) -> str:
    """Get the type render for a type ID."""

    render_url = type_render_url(type_id=type_id, size=size)

    if as_html:
        render_html = format_html(
            '<img class="type-render rounded-circle" src="{}" alt="{}">',
            render_url,
            type_name,
        )
        return render_html
    return render_url


def get_badge_html(label: str, color: str = "primary", size: str = "sm") -> str:
    """Get a badge HTML element."""

    return format_html(
        '<span class="badge badge-{} badge-{}">{}</span>',
        color,
        size,
        label,
    )


def str_normalize_time(evetime, hours: bool = False) -> str:
    """Normalize time to a string."""
    if hours:
        return timezone.localtime(evetime).strftime("%Y-%m-%d %H:%M")
    return timezone.localtime(evetime).strftime("%Y-%m-%d")


def generate_icon(
    color: str, icon: str, size: str = 32, title: str = None, position: str = "end"
) -> str:
    """Generate a bootstrap icon button."""
    html = f"<div class='d-flex justify-content-{position}'>"
    html += f"<button class='btn btn-{color} btn-sm d-flex align-items-center justify-content-center'"
    html += f" style='height: {size}px; width: {size}px;' data-tooltip-toggle='taxsystem-tooltip'"
    if title:
        html += f" title='{title}'"
    html += ">"
    html += f"<i class='{icon}'></i>"
    html += "</button>"
    html += "</div>"
    return format_html(html)



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.utils import timezone

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA TaxSystem
from taxsystem import __title__
from taxsystem.decorators import log_timing

if TYPE_CHECKING:
    # AA TaxSystem
    from taxsystem.models.tax import OwnerAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class PaymentSystemQuerySet(models.QuerySet):
    pass


class PaymentSystemManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_payment_system(
        self, owner: "OwnerAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create Payment System data."""
        return owner.update_section_if_changed(
            section=owner.UpdateSection.PAYMENT_SYSTEM,
            fetch_func=self._update_or_create_objs,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=unused-argument
    def _update_or_create_objs(
        self, owner: "OwnerAudit", force_refresh: bool = False, runs: int = 0
    ) -> None:
        """Update or Create payment system entries from objs data."""
        # pylint: disable=import-outside-toplevel, cyclic-import
        # AA TaxSystem
        from taxsystem.models.filters import JournalFilterSet
        from taxsystem.models.logs import PaymentHistory
        from taxsystem.models.tax import Payments

        logger.debug(
            "Updating Payment System for: %s",
            owner.corporation.corporation_name,
        )

        payments = Payments.objects.filter(
            account__owner=owner, request_status=Payments.RequestStatus.PENDING
        )

        _current_payment_ids = set(payments.values_list("id", flat=True))
        _automatic_payment_ids = []

        # Check for any automatic payments
        try:
            filters_obj = JournalFilterSet.objects.filter(owner=owner)
            for filter_obj in filters_obj:
                payments = filter_obj.filter(payments)
                for payment in payments:
                    if payment.request_status == Payments.RequestStatus.PENDING:
                        # Ensure all transfers are processed in a single transaction
                        with transaction.atomic():
                            payment.request_status = Payments.RequestStatus.APPROVED
                            payment.reviser = "System"

                            # Update payment pool for user
                            self.filter(owner=owner, user=payment.account.user).update(
                                deposit=payment.account.deposit + payment.amount
                            )

                            payment.save()

                            PaymentHistory(
                                user=payment.account.user,
                                payment=payment,
                                action=PaymentHistory.Actions.STATUS_CHANGE,
                                new_status=Payments.RequestStatus.APPROVED,
                                comment=PaymentHistory.SystemText.AUTOMATIC,
                            ).save()

                            runs = runs + 1
                            _automatic_payment_ids.append(payment.pk)
        except JournalFilterSet.DoesNotExist:
            pass

        # Check for any payments that need approval
        needs_approval = _current_payment_ids - set(_automatic_payment_ids)
        approvals = Payments.objects.filter(
            id__in=needs_approval, request_status=Payments.RequestStatus.PENDING
        )

        for payment in approvals:
            payment.request_status = Payments.RequestStatus.NEEDS_APPROVAL
            payment.save()

            PaymentHistory(
                user=payment.account.user,
                payment=payment,
                action=PaymentHistory.Actions.STATUS_CHANGE,
                new_status=Payments.RequestStatus.NEEDS_APPROVAL,
                comment=PaymentHistory.SystemText.REVISER,
            ).save()

            runs = runs + 1

        logger.debug(
            "Finished %s: Payment System entrys for %s",
            runs,
            owner.corporation.corporation_name,
        )

        return ("Finished Payment System for %s", owner.corporation.corporation_name)

    @log_timing(logger)
    def check_pay_day(self, owner: "OwnerAudit", force_refresh: bool = False) -> None:
        """Update or Create Payment System data."""
        return owner.update_section_if_changed(
            section=owner.UpdateSection.PAYDAY,
            fetch_func=self._pay_day,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=unused-argument
    def _pay_day(
        self, owner: "OwnerAudit", force_refresh: bool = False, runs: int = 0
    ) -> None:
        """Update or Create payment system entries from objs data."""
        logger.debug(
            "Updating payday for: %s",
            owner.corporation.corporation_name,
        )

        payment_system = self.filter(owner=owner, status=self.model.Status.ACTIVE)

        for user in payment_system:
            if user.last_paid is None:
                # First Period is free
                user.last_paid = timezone.now()
            if timezone.now() - user.last_paid >= timezone.timedelta(
                days=owner.tax_period
            ):
                user.deposit -= owner.tax_amount
                user.last_paid = timezone.now()
                runs = runs + 1
            user.save()

        logger.debug(
            "Finished %s: Payday for %s",
            runs,
            owner.corporation.corporation_name,
        )

        return ("Finished Payday for %s", owner.corporation.corporation_name)


PaymentSystemManager = PaymentSystemManagerBase.from_queryset(PaymentSystemQuerySet)


class PaymentsQuerySet(models.QuerySet):
    pass


class PaymentsManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_payments(
        self, owner: "OwnerAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create a Payments entry data."""
        return owner.update_section_if_changed(
            section=owner.UpdateSection.PAYMENTS,
            fetch_func=self._update_or_create_objs,
            force_refresh=force_refresh,
        )

    @transaction.atomic()
    # pylint: disable=too-many-locals, unused-argument
    def _update_or_create_objs(
        self, owner: "OwnerAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create payment system entries from objs data."""
        # pylint: disable=import-outside-toplevel, cyclic-import
        # AA TaxSystem
        from taxsystem.models.logs import PaymentHistory
        from taxsystem.models.tax import Payments, PaymentSystem
        from taxsystem.models.wallet import CorporationWalletJournalEntry

        logger.debug(
            "Updating payments for: %s",
            owner.corporation.corporation_name,
        )

        accounts = PaymentSystem.objects.filter(owner=owner)

        if not accounts:
            return ("No Payment Users for %s", owner.corporation.corporation_name)

        users = {}

        for user in accounts:
            user: PaymentSystem
            alts = user.get_alt_ids()
            users[user] = alts

        journal = CorporationWalletJournalEntry.objects.filter(
            division__corporation=owner, ref_type__in=["player_donation"]
        ).order_by("-date")

        _current_entry_ids = set(
            self.filter(account__owner=owner).values_list("entry_id", flat=True)
        )
        with transaction.atomic():
            items = []
            logs_items = []
            for entry in journal:
                # Skip if already processed
                if entry.entry_id in _current_entry_ids:
                    continue
                for user, alts in users.items():
                    if entry.first_party.id in alts:
                        payment_item = Payments(
                            entry_id=entry.entry_id,
                            name=user.name,
                            account=user,
                            amount=entry.amount,
                            request_status=Payments.RequestStatus.PENDING,
                            date=entry.date,
                            reason=entry.reason,
                        )
                        items.append(payment_item)

            payments = self.bulk_create(items, ignore_conflicts=True)

            for payment in payments:
                try:
                    payment = self.get(
                        entry_id=payment.entry_id, account=payment.account
                    )
                except Payments.DoesNotExist:
                    continue

                log_items = PaymentHistory(
                    user=payment.account.user,
                    payment=payment,
                    action=PaymentHistory.Actions.STATUS_CHANGE,
                    new_status=Payments.RequestStatus.PENDING,
                    comment=PaymentHistory.SystemText.ADDED,
                )
                logs_items.append(log_items)

            PaymentHistory.objects.bulk_create(logs_items, ignore_conflicts=True)

        logger.debug(
            "Finished %s Payments for %s",
            len(items),
            owner.corporation.corporation_name,
        )
        return (
            "Finished %s Payments for %s",
            len(items),
            owner.corporation.corporation_name,
        )


PaymentsManager = PaymentsManagerBase.from_queryset(PaymentsQuerySet)



# Django
from django.test import TestCase
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import create_user_from_evecharacter
from eveuniverse.models import EveEntity

# AA TaxSystem
from taxsystem.models.tax import Payments
from taxsystem.tests.testdata.generate_owneraudit import (
    create_owneraudit_from_user,
)
from taxsystem.tests.testdata.generate_payments import (
    create_payment,
    create_payment_system,
)
from taxsystem.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from taxsystem.tests.testdata.load_allianceauth import load_allianceauth
from taxsystem.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "taxsystem.models.tax"


class TestPaymentsModel(TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001, permissions=["taxsystem.basic_access"]
        )
        cls.user2, cls.character_ownership2 = create_user_from_evecharacter(
            1002, permissions=["taxsystem.basic_access"]
        )
        cls.audit = create_owneraudit_from_user(cls.user)
        cls.audit2 = create_owneraudit_from_user(cls.user2)

        cls.eve_character_first_party = EveEntity.objects.get(id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(id=1001)

        cls.division = create_division(
            corporation=cls.audit,
            division_id=1,
            name="Main Division",
            balance=1000000,
        )

        cls.journal_entry = create_wallet_journal_entry(
            division=cls.division,
            entry_id=1,
            amount=1000,
            date=timezone.datetime(2025, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            reason="Test Journal Entry",
            ref_type="tax_payment",
            first_party=cls.eve_character_first_party,
            second_party=cls.eve_character_second_party,
        )

        cls.payment_system = create_payment_system(
            name=cls.character_ownership.character.character_name,
            owner=cls.audit,
            user=cls.user,
        )

        cls.payments = create_payment(
            name="Gneuten",
            entry_id=1,
            account=cls.payment_system,
            amount=1000,
            date=timezone.datetime(2025, 1, 1, 12, 0, 0, tzinfo=timezone.utc),
            request_status="needs_approval",
            reviser="Requires Auditor",
        )

    def test_str(self):
        expected_str = Payments.objects.get(account=self.payment_system)
        self.assertEqual(self.payments, expected_str)

    def test_is_automatic(self):
        """Test if the payment is automatic."""
        payments = Payments.objects.get(account=self.payment_system)
        self.assertFalse(payments.is_automatic)

    def test_is_pending(self):
        """Test if the payment is pending."""
        self.payments.request_status = Payments.RequestStatus.PENDING
        self.payments.save()

        payments = Payments.objects.get(account=self.payment_system)
        self.assertTrue(payments.is_pending)

    def test_is_approved(self):
        """Test if the payment is approved."""
        self.payments.request_status = Payments.RequestStatus.APPROVED
        self.payments.save()

        payments = Payments.objects.get(account=self.payment_system)
        self.assertFalse(payments.is_pending)
        self.assertTrue(payments.is_approved)

    def test_is_rejected(self):
        """Test if the payment is rejected."""
        self.payments.request_status = Payments.RequestStatus.REJECTED
        self.payments.save()

        payments = Payments.objects.get(account=self.payment_system)
        self.assertFalse(payments.is_pending)
        self.assertTrue(payments.is_rejected)

    def test_character_id(self):
        """Test if the character_id is correct."""
        payments = Payments.objects.get(account=self.payment_system)
        self.assertEqual(
            payments.character_id, self.character_ownership.character.character_id
        )

    def test_division(self):
        """Test if the division is correct."""
        payments = Payments.objects.get(account=self.payment_system)
        self.assertEqual(payments.division, "Main Division")



"""
App Settings
"""

# Standard Library
import sys

# Alliance Auth (External Libs)
from app_utils.app_settings import clean_setting

IS_TESTING = sys.argv[1:2] == ["test"]

# EVE Online Swagger
EVE_BASE_URL = "https://esi.evetech.net/"
EVE_API_URL = "https://esi.evetech.net/latest/"
EVE_BASE_URL_REGEX = r"^http[s]?:\/\/esi.evetech\.net\/"

# Fuzzwork
FUZZ_BASE_URL = "https://www.fuzzwork.co.uk/"
FUZZ_API_URL = "https://www.fuzzwork.co.uk/api/"
FUZZ_BASE_URL_REGEX = r"^http[s]?:\/\/(www\.)?fuzzwork\.co\.uk\/"

# ZKillboard
ZKILLBOARD_BASE_URL = "https://zkillboard.com/"
ZKILLBOARD_API_URL = "https://zkillboard.com/api/"
ZKILLBOARD_BASE_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/"
ZKILLBOARD_KILLMAIL_URL_REGEX = r"^http[s]?:\/\/zkillboard\.com\/kill\/\d+\/"

# Set Naming on Auth Hook
TAXSYSTEM_APP_NAME = clean_setting("TAXSYSTEM_APP_NAME", "Tax System")

# Task Settings
# Global timeout for tasks in seconds to reduce task accumulation during outages.
TAXSYSTEM_TASKS_TIME_LIMIT = clean_setting("TAXSYSTEM_TASKS_TIME_LIMIT", 7200)

# Stale time in minutes for each type of data
TAXSYSTEM_STALE_TYPES = {
    "wallet": 60,
    "division_names": 60,
    "division": 15,
    "members": 60,
    "payments": 60,
    "payment_system": 60,
    "payment_payday": 1440,
}



# Django
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.api.helpers import generate_button
from taxsystem.models.filters import JournalFilter


def _filter_actions(corporation_id, filter_obj: JournalFilter, perms, request):
    action = ""
    if perms:
        action = generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=filter_obj,
            settings={
                "title": _("Delete Filter"),
                "icon": "fas fa-trash",
                "color": "danger",
                "text": _("Delete Filter"),
                "modal": "filter-set-delete-filter",
                "action": reverse(
                    viewname="taxsystem:delete_filter",
                    kwargs={
                        "corporation_id": corporation_id,
                        "filter_pk": filter_obj.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )

    actions_html = format_html(action)
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)



"""Hook into Alliance Auth"""

# Standard Library
import logging

# Django
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth import hooks
from allianceauth.services.hooks import MenuItemHook, UrlHook

# AA TaxSystem
from taxsystem.models.tax import PaymentSystem

from . import app_settings, urls

logger = logging.getLogger(__name__)


class TaxSystemMenuItem(MenuItemHook):
    """This class ensures only authorized users will see the menu entry"""

    def __init__(self):
        super().__init__(
            f"{app_settings.TAXSYSTEM_APP_NAME}",
            "fas fa-landmark fa-fw",
            "taxsystem:index",
            navactive=["taxsystem:"],
        )

    def render(self, request):
        if request.user.has_perm("taxsystem.basic_access"):
            try:
                payment_user = PaymentSystem.objects.get(user=request.user)
                self.count = 1 if not payment_user.has_paid else 0
            except PaymentSystem.DoesNotExist:
                self.count = 0
            return MenuItemHook.render(self, request)
        return ""


@hooks.register("menu_item_hook")
def register_menu():
    """Register the menu item"""

    return TaxSystemMenuItem()


@hooks.register("url_hook")
def register_urls():
    """Register app urls"""

    return UrlHook(urls, "taxsystem", r"^taxsystem/")



# Alliance Auth
from allianceauth.authentication.models import UserProfile


# pylint: disable=unused-argument
def add_info_to_context(request, context: dict) -> dict:
    """Add additional information to the context for the view."""
    theme = None
    try:
        user = UserProfile.objects.get(id=request.user.id)
        theme = user.theme
    except UserProfile.DoesNotExist:
        pass

    new_context = {
        **{"theme": theme},
        **context,
    }
    return new_context



# Standard Library
from unittest.mock import MagicMock, patch

# Django
from django.test import override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter
from eveuniverse.models import EveEntity

# AA TaxSystem
from taxsystem.tests.testdata.esi_stub import esi_client_stub
from taxsystem.tests.testdata.generate_owneraudit import create_owneraudit_from_user
from taxsystem.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from taxsystem.tests.testdata.load_allianceauth import load_allianceauth
from taxsystem.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "taxsystem.managers.wallet_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".etag_results")
@patch(MODULE_PATH + ".EveEntity.objects.bulk_resolve_ids")
@patch(MODULE_PATH + ".EveEntity.objects.filter")
class TestWalletManager(NoSocketsTestCase):
    """Test Wallet Manager for Corporation Journal Entries."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
        )
        cls.audit = create_owneraudit_from_user(cls.user)

        cls.eve_character_first_party = EveEntity.objects.get(id=2001)
        cls.eve_character_second_party = EveEntity.objects.get(id=1001)

        cls.division = create_division(
            corporation=cls.audit, name="MEGA KONTO", balance=1000000, division_id=1
        )
        cls.token = cls.character_ownership.user.token_set.first()
        cls.audit.get_token = MagicMock(return_value=cls.token)

    def test_update_wallet_journal(
        self, mock_filter, mock_entity_bulk, mock_etag, mock_esi
    ):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()[0]
        filter_mock = mock_filter.return_value
        filter_mock.count.return_value = 0

        mock_entity_bulk.side_effect = [
            EveEntity.objects.create(
                id=9998,
                name="Test Character",
                category="character",
            )
        ]

        self.audit.update_wallet(force_refresh=False)

        self.assertSetEqual(
            set(self.division.ts_corporation_wallet.values_list("entry_id", flat=True)),
            {10, 13, 16},
        )
        obj = self.division.ts_corporation_wallet.get(entry_id=10)
        self.assertEqual(obj.amount, 1000)
        self.assertEqual(obj.context_id, 1)
        self.assertEqual(obj.first_party.id, 2001)
        self.assertEqual(obj.second_party.id, 1001)

        obj = self.division.ts_corporation_wallet.get(entry_id=13)
        self.assertEqual(obj.amount, 5000)

        obj = self.division.ts_corporation_wallet.get(entry_id=16)
        self.assertEqual(obj.amount, 10000)



# Standard Library
from unittest.mock import patch

# Django
from django.test import override_settings
from django.utils import timezone

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase, create_user_from_evecharacter
from eveuniverse.models import EveEntity

# AA TaxSystem
from taxsystem.tests.testdata.esi_stub import esi_client_stub
from taxsystem.tests.testdata.generate_owneraudit import create_owneraudit_from_user
from taxsystem.tests.testdata.generate_walletjournal import (
    create_division,
    create_wallet_journal_entry,
)
from taxsystem.tests.testdata.load_allianceauth import load_allianceauth
from taxsystem.tests.testdata.load_eveuniverse import load_eveuniverse

MODULE_PATH = "taxsystem.managers.wallet_manager"


@override_settings(CELERY_ALWAYS_EAGER=True, CELERY_EAGER_PROPAGATES_EXCEPTIONS=True)
@patch(MODULE_PATH + ".esi")
@patch(MODULE_PATH + ".etag_results")
class TestDivisionManager(NoSocketsTestCase):
    """Test Division Manager for Corporation Divisions."""

    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        load_allianceauth()
        load_eveuniverse()

        cls.user, cls.character_ownership = create_user_from_evecharacter(
            1001,
        )
        cls.audit = create_owneraudit_from_user(cls.user)

    def test_update_division_names(self, mock_etag, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()

        self.audit.update_division_names(force_refresh=False)

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=2
        )
        self.assertEqual(obj.name, "Rechnungen")

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=4
        )
        self.assertEqual(obj.name, "Ship Replacment Abteilung")

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=6
        )
        self.assertEqual(obj.name, "Partner")

    def test_update_division(self, mock_etag, mock_esi):
        # given
        mock_esi.client = esi_client_stub
        mock_etag.side_effect = lambda ob, token, force_refresh=False: ob.results()

        self.audit.update_division(force_refresh=False)

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=2
        )
        self.assertEqual(obj.balance, 0)

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=4
        )
        self.assertEqual(obj.balance, 1600000000)

        obj = self.audit.ts_corporation_division.get(
            corporation__corporation__corporation_id=2001, division_id=6
        )
        self.assertEqual(obj.balance, 0)



# Standard Library
import logging

# Third Party
from ninja import NinjaAPI

# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.db.models import Sum
from django.shortcuts import render
from django.utils import timezone
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.api.helpers import get_manage_corporation
from taxsystem.api.taxsystem.helpers.administration import _delete_member
from taxsystem.api.taxsystem.helpers.filters import _filter_actions
from taxsystem.api.taxsystem.helpers.payments import payments_actions
from taxsystem.api.taxsystem.helpers.paymentsystem import (
    _payments_info,
    payment_system_actions,
)
from taxsystem.api.taxsystem.helpers.statistics import (
    _get_divisions_dict,
    _get_statistics_dict,
)
from taxsystem.helpers import lazy
from taxsystem.models.filters import JournalFilter
from taxsystem.models.logs import AdminLogs
from taxsystem.models.tax import Members, Payments, PaymentSystem
from taxsystem.models.wallet import (
    CorporationWalletDivision,
    CorporationWalletJournalEntry,
)

logger = logging.getLogger(__name__)


class AdminApiEndpoints:
    tags = ["Admin"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/view/dashboard/",
            response={200: dict, 403: str, 404: str},
            tags=self.tags,
        )
        # pylint: disable=too-many-locals
        def get_dashboard(request, corporation_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            divisions = CorporationWalletDivision.objects.filter(corporation=owner)

            corporation_id = owner.corporation.corporation_id
            corporation_logo = lazy.get_corporation_logo_url(
                corporation_id, size=64, as_html=True
            )
            divisions_dict = _get_divisions_dict(divisions)
            statistics_dict = {owner.name: _get_statistics_dict(owner)}

            past30_days = (
                CorporationWalletJournalEntry.objects.filter(
                    division__corporation=owner,
                    date__gte=timezone.now() - timezone.timedelta(days=30),
                )
                .exclude(first_party_id=corporation_id, second_party_id=corporation_id)
                .aggregate(total=Sum("amount"))
            )

            total_amount = past30_days.get("total", 0) or 0
            activity_color = "text-success" if total_amount >= 0 else "text-danger"
            activity_html = f"<span class='{activity_color}'>{intcomma(total_amount, use_l10n=True)}</span> ISK"

            output = {
                "corporation_name": owner.name,
                "corporation_id": corporation_id,
                "corporation_logo": corporation_logo,
                "update_status_icon": owner.get_status.bootstrap_icon(),
                "update_status": owner.get_update_status,
                "tax_amount": owner.tax_amount,
                "tax_period": owner.tax_period,
                "divisions": divisions_dict,
                "statistics": statistics_dict,
                "activity": format_html(activity_html),
            }

            return output

        @api.get(
            "corporation/{corporation_id}/view/members/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_members(request, corporation_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            corporation_dict = {}

            members = Members.objects.filter(owner=owner)

            for member in members:
                actions = _delete_member(
                    corporation_id=corporation_id,
                    member=member,
                    perms=perms,
                    request=request,
                )

                corporation_dict[member.character_id] = {
                    "character_id": member.character_id,
                    "character_portrait": lazy.get_character_portrait_url(
                        member.character_id, size=32, as_html=True
                    ),
                    "character_name": member.character_name,
                    "is_faulty": member.is_faulty,
                    "status": member.get_status_display(),
                    "joined": lazy.str_normalize_time(member.joined, hours=False),
                    "actions": actions,
                }

            output = []
            output.append({"corporation": corporation_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/view/paymentsystem/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_paymentsystem(request, corporation_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            payment_system = (
                PaymentSystem.objects.filter(
                    owner=owner,
                    user__profile__main_character__isnull=False,
                )
                .exclude(status=PaymentSystem.Status.MISSING)
                .select_related(
                    "user", "user__profile", "user__profile__main_character"
                )
            )

            payment_dict = {}

            for user in payment_system:
                character_id = user.user.profile.main_character.character_id
                character_name = user.user.profile.main_character.character_name

                actions = payment_system_actions(
                    corporation_id=corporation_id,
                    payment_system=user,
                    perms=perms,
                    request=request,
                )
                deposit = _payments_info(
                    corporation_id=corporation_id,
                    user=user,
                    perms=perms,
                    request=request,
                )
                payment_dict[character_id] = {
                    "character_id": character_id,
                    "character_portrait": lazy.get_character_portrait_url(
                        character_id=character_id,
                        size=32,
                        as_html=True,
                    ),
                    "character_name": character_name,
                    "alts": user.get_alt_ids(),
                    "status": user.get_payment_status(),
                    "deposit": deposit,
                    "has_paid": user.has_paid_icon(badge=True),
                    "has_paid_filter": _("Yes") if user.has_paid else _("No"),
                    "last_paid": lazy.str_normalize_time(user.last_paid, hours=True),
                    "is_active": user.is_active,
                    "actions": actions,
                }

            output = []
            output.append({"corporation": payment_dict})

            return output

        @api.get(
            "corporation/admin/{corporation_id}/view/logs/",
            response={200: list, 403: str},
            tags=self.tags,
        )
        def get_corporation_admin_logs(request, corporation_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            logs = AdminLogs.objects.filter(owner=owner).order_by("-date")

            logs_dict = {}

            for log in logs:
                date = lazy.str_normalize_time(log.date, hours=True)
                logs_dict[log.pk] = {
                    "date": date,
                    "user_name": log.user.username,
                    "action": log.action,
                    "log": log.log,
                }

            output = []
            output.append({"logs": logs_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/character/{character_id}/view/payments/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_main_character_payments(
            request, corporation_id: int, character_id: int
        ):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            payments = Payments.objects.filter(
                account__owner=owner,
                account__user__profile__main_character__character_id=character_id,
                corporation_id=owner.corporation.corporation_id,
            )

            if not payments:
                return 404, "No Payments Found"

            # Create a dict for the character
            payments_char_dict = {
                "title": "Payments for",
                "character_id": character_id,
                "character_portrait": lazy.get_character_portrait_url(
                    character_id, size=32, as_html=True
                ),
                "character_name": payments[0].account.name,
            }

            # Create a dict for each payment
            payments_dict = {}
            for payment in payments:
                try:
                    character_id = (
                        payment.account.user.profile.main_character.character_id
                    )
                    portrait = lazy.get_character_portrait_url(
                        character_id, size=32, as_html=True
                    )
                except AttributeError:
                    portrait = ""

                actions = payments_actions(corporation_id, payment, perms, request)
                amount = f"{intcomma(payment.amount, use_l10n=True)} ISK"

                payments_dict[payment.pk] = {
                    "payment_id": payment.pk,
                    "character_portrait": portrait,
                    "character_name": payment.account.name,
                    "payment_date": payment.formatted_payment_date,
                    "amount": amount,
                    "request_status": payment.get_request_status_display(),
                    "reviser": payment.reviser,
                    "division": payment.division,
                    "reason": payment.reason,
                    "actions": actions,
                }

            # Add payments to the character dict
            payments_char_dict["payments"] = payments_dict

            context = {
                "entity_pk": corporation_id,
                "entity_type": "character",
                "character": payments_char_dict,
            }

            return render(
                request,
                "taxsystem/modals/view_character_payments.html",
                context,
            )

        @api.get(
            "corporation/{corporation_id}/filter_set/{filter_set_id}/view/filter/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_filter_set_filters(request, corporation_id: int, filter_set_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            filters = JournalFilter.objects.filter(
                filter_set__pk=filter_set_id,
            )

            output = []

            for filter_obj in filters:
                if filter_obj.filter_type == JournalFilter.FilterType.AMOUNT:
                    value = f"{intcomma(filter_obj.value, use_l10n=True)} ISK"
                else:
                    value = filter_obj.value

                filter_dict = {
                    "id": filter_obj.pk,
                    "filter_set": filter_obj.filter_set,
                    "filter_type": filter_obj.get_filter_type_display(),
                    "value": value,
                    "actions": _filter_actions(
                        corporation_id=corporation_id,
                        filter_obj=filter_obj,
                        perms=perms,
                        request=request,
                    ),
                }
                output.append(filter_dict)

            return render(
                request,
                "taxsystem/modals/view_filter.html",
                context={
                    "filters": output,
                },
            )



# Django
from django.contrib.auth.models import User

# Alliance Auth
from allianceauth.authentication.backends import StateBackend
from allianceauth.authentication.models import CharacterOwnership
from allianceauth.eveonline.models import EveCharacter
from allianceauth.tests.auth_utils import AuthUtils

# Alliance Auth (External Libs)
from app_utils.testing import add_character_to_user

# AA TaxSystem
# AA Taxsystem
from taxsystem.models.tax import OwnerAudit, OwnerUpdateStatus


def create_character(eve_character: EveCharacter, **kwargs) -> OwnerAudit:
    """Create a Taxsystem Character from EveCharacter"""
    params = {
        "name": eve_character.corporation.corporation_name,
        "corporation": eve_character.corporation,
    }
    params.update(kwargs)
    corporation = OwnerAudit(**params)
    corporation.save()
    return corporation


def create_update_status(owner_audit: OwnerAudit, **kwargs) -> OwnerUpdateStatus:
    """Create a Update Status for a Character Audit"""
    params = {
        "owner": owner_audit,
    }
    params.update(kwargs)
    update_status = OwnerUpdateStatus(**params)
    update_status.save()
    return update_status


def create_owneraudit_from_user(user: User, **kwargs) -> OwnerAudit:
    """Create a Character Audit from a user"""
    eve_character = user.profile.main_character
    if not eve_character:
        raise ValueError("User needs to have a main character.")

    kwargs.update({"eve_character": eve_character})
    return create_character(**kwargs)


def create_user_from_evecharacter_with_access(
    character_id: int, disconnect_signals: bool = True
) -> tuple[User, CharacterOwnership]:
    """Create user with access from an existing eve character and use it as main."""
    auth_character = EveCharacter.objects.get(character_id=character_id)
    username = StateBackend.iterate_username(auth_character.character_name)
    user = AuthUtils.create_user(username, disconnect_signals=disconnect_signals)
    user = AuthUtils.add_permission_to_user_by_name(
        "taxsystem.basic_access", user, disconnect_signals=disconnect_signals
    )
    character_ownership = add_character_to_user(
        user,
        auth_character,
        is_main=True,
        scopes=OwnerAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )
    return user, character_ownership


def create_owneraudit_from_evecharacter(character_id: int, **kwargs) -> OwnerAudit:
    """Create a Audit Character from a existing EveCharacter"""

    _, character_ownership = create_user_from_evecharacter_with_access(
        character_id, disconnect_signals=True
    )
    return create_character(character_ownership.character, **kwargs)


def add_auth_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True
) -> CharacterOwnership:
    auth_character = EveCharacter.objects.get(character_id=character_id)
    return add_character_to_user(
        user,
        auth_character,
        is_main=False,
        scopes=OwnerAudit.get_esi_scopes(),
        disconnect_signals=disconnect_signals,
    )


def add_owneraudit_character_to_user(
    user: User, character_id: int, disconnect_signals: bool = True, **kwargs
) -> OwnerAudit:
    """Add a Character Audit Character to a user"""
    character_ownership = add_auth_character_to_user(
        user,
        character_id,
        disconnect_signals=disconnect_signals,
    )
    return create_character(character_ownership.character, **kwargs)



# Django
from django.db import models

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA TaxSystem
from taxsystem import __title__

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class LogsQuerySet(models.QuerySet):
    pass


class LogsManagerBase(models.Manager):
    pass


LogsManager = LogsManagerBase.from_queryset(LogsQuerySet)



"""Models for Tax System."""

# Standard Library
from collections.abc import Callable
from typing import TYPE_CHECKING

# Third Party
from bravado.exception import HTTPInternalServerError

# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.utils import timezone
from django.utils.safestring import mark_safe
from django.utils.translation import gettext_lazy as _

# Alliance Auth
from allianceauth.authentication.models import OwnershipRecord, User
from allianceauth.eveonline.models import (
    EveAllianceInfo,
    EveCharacter,
    EveCorporationInfo,
)
from allianceauth.services.hooks import get_extension_logger
from esi.errors import TokenError
from esi.models import Token

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA TaxSystem
from taxsystem import __title__, app_settings
from taxsystem.errors import HTTPGatewayTimeoutError, NotModifiedError
from taxsystem.managers.payment_manager import PaymentsManager, PaymentSystemManager
from taxsystem.managers.tax_manager import MembersManager, OwnerAuditManager
from taxsystem.models.general import UpdateSectionResult, _NeedsUpdate
from taxsystem.providers import esi

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

if TYPE_CHECKING:
    # AA TaxSystem
    from taxsystem.models.wallet import CorporationWalletJournalEntry


class OwnerAudit(models.Model):
    """Tax System Audit model for app"""

    class UpdateSection(models.TextChoices):
        WALLET = "wallet", _("Wallet Journal")
        DIVISION_NAMES = "division_names", _("Wallet Division Names")
        DIVISION = "division", _("Wallet Division")
        MEMBERS = "members", _("Members")
        PAYMENTS = "payments", _("Payments")
        PAYMENT_SYSTEM = "payment_system", _("Payment System")
        PAYDAY = "payday", _("Payday")

        @classmethod
        def get_sections(cls) -> list[str]:
            """Return list of section values."""
            return [choice.value for choice in cls]

        @property
        def method_name(self) -> str:
            """Return method name for this section."""
            return f"update_{self.value}"

    class UpdateStatus(models.TextChoices):
        DISABLED = "disabled", _("disabled")
        TOKEN_ERROR = "token_error", _("token error")
        ERROR = "error", _("error")
        OK = "ok", _("ok")
        INCOMPLETE = "incomplete", _("incomplete")
        IN_PROGRESS = "in_progress", _("in progress")

        def bootstrap_icon(self) -> str:
            """Return bootstrap corresponding icon class."""
            update_map = {
                status: mark_safe(
                    f"<span class='{self.bootstrap_text_style_class()}' data-tooltip-toggle='taxsystem-tooltip' title='{self.description()}'>⬤</span>"
                )
                for status in [
                    self.DISABLED,
                    self.TOKEN_ERROR,
                    self.ERROR,
                    self.INCOMPLETE,
                    self.IN_PROGRESS,
                    self.OK,
                ]
            }
            return update_map.get(self, "")

        def bootstrap_text_style_class(self) -> str:
            """Return bootstrap corresponding bootstrap text style class."""
            update_map = {
                self.DISABLED: "text-muted",
                self.TOKEN_ERROR: "text-warning",
                self.INCOMPLETE: "text-warning",
                self.IN_PROGRESS: "text-info",
                self.ERROR: "text-danger",
                self.OK: "text-success",
            }
            return update_map.get(self, "")

        def description(self) -> str:
            """Return description for an enum object."""
            update_map = {
                self.DISABLED: _("Update is disabled"),
                self.TOKEN_ERROR: _("One section has a token error during update"),
                self.INCOMPLETE: _("One or more sections have not been updated"),
                self.IN_PROGRESS: _("Update is in progress"),
                self.ERROR: _("An error occurred during update"),
                self.OK: _("Updates completed successfully"),
            }
            return update_map.get(self, "")

    objects = OwnerAuditManager()

    name = models.CharField(
        max_length=255,
    )

    corporation = models.OneToOneField(
        EveCorporationInfo, on_delete=models.CASCADE, related_name="+"
    )

    alliance = models.ForeignKey(
        EveAllianceInfo,
        on_delete=models.CASCADE,
        related_name="alliance",
        blank=True,
        null=True,
    )

    active = models.BooleanField(default=True)

    tax_amount = models.DecimalField(
        max_digits=16,
        decimal_places=0,
        help_text=_("Tax Amount in ISK that is set for the corporation. Max 16 Digits"),
        default=0,
        validators=[MaxValueValidator(9999999999999999)],
    )

    tax_period = models.PositiveIntegerField(
        help_text=_(
            "Tax Period in days for the corporation. Max 365 days. Default: 30 days"
        ),
        default=30,
        validators=[MaxValueValidator(365)],
    )

    def __str__(self):
        return f"{self.corporation.corporation_name} - Status: {self.get_status}"

    def update_division_names(self, force_refresh: bool) -> None:
        """Update the divisions for this corporation."""
        return self.ts_corporation_division.update_or_create_esi_names(
            self, force_refresh=force_refresh
        )

    def update_division(self, force_refresh: bool) -> None:
        """Update the divisions for this corporation."""
        return self.ts_corporation_division.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def update_wallet(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the wallet journal for this corporation."""
        # pylint: disable=import-outside-toplevel
        # AA TaxSystem
        from taxsystem.models.wallet import CorporationWalletJournalEntry

        return CorporationWalletJournalEntry.objects.update_or_create_esi(
            self, force_refresh=force_refresh
        )

    def update_members(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the members for this corporation."""
        return self.ts_members.update_or_create_esi(self, force_refresh=force_refresh)

    def update_payments(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the Payments for this corporation."""
        return Payments.objects.update_or_create_payments(
            self, force_refresh=force_refresh
        )

    def update_payment_system(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the Payment System for this corporation."""
        return self.ts_payment_system.update_or_create_payment_system(
            self, force_refresh=force_refresh
        )

    def update_payday(self, force_refresh: bool) -> UpdateSectionResult:
        """Update the Payment System for this corporation."""
        return self.ts_payment_system.check_pay_day(self, force_refresh=force_refresh)

    @classmethod
    def get_esi_scopes(cls) -> list[str]:
        """Return list of required ESI scopes to fetch."""
        return [
            # General
            "esi-corporations.read_corporation_membership.v1",
            "esi-corporations.track_members.v1",
            "esi-characters.read_corporation_roles.v1",
            # wallets
            "esi-wallet.read_corporation_wallets.v1",
            "esi-corporations.read_divisions.v1",
        ]

    def get_token(self, scopes, req_roles) -> Token:
        """Get the token for this corporation."""
        if "esi-characters.read_corporation_roles.v1" not in scopes:
            scopes.append("esi-characters.read_corporation_roles.v1")

        char_ids = EveCharacter.objects.filter(
            corporation_id=self.corporation.corporation_id
        ).values("character_id")

        tokens = Token.objects.filter(character_id__in=char_ids).require_scopes(scopes)

        for token in tokens:
            try:
                roles = esi.client.Character.get_characters_character_id_roles(
                    character_id=token.character_id, token=token.valid_access_token()
                ).result()

                has_roles = False
                for role in roles.get("roles", []):
                    if role in req_roles:
                        has_roles = True

                if has_roles:
                    return token
            except TokenError as e:
                logger.error(
                    "Token ID: %s (%s)",
                    token.pk,
                    e,
                )
        return False

    @property
    def get_status(self) -> UpdateStatus:
        """Get the status of this character."""
        if self.active is False:
            return self.UpdateStatus.DISABLED

        qs = OwnerAudit.objects.filter(pk=self.pk).annotate_total_update_status()
        total_update_status = list(qs.values_list("total_update_status", flat=True))[0]
        return self.UpdateStatus(total_update_status)

    @property
    def get_update_status(self) -> dict[str, str]:
        """Return a dictionary of update sections and their statuses."""
        update_status = {}
        for section in self.UpdateSection.get_sections():
            try:
                status = self.ts_update_status.get(section=section)
                update_status[section] = {
                    "is_success": status.is_success,
                    "last_update_finished_at": status.last_update_finished_at,
                    "last_run_finished_at": status.last_run_finished_at,
                }
            except OwnerUpdateStatus.DoesNotExist:
                continue
        return update_status

    def calc_update_needed(self) -> _NeedsUpdate:
        """Calculate if an update is needed."""
        sections_needs_update = {
            section: True for section in self.UpdateSection.get_sections()
        }
        existing_sections: models.QuerySet[OwnerUpdateStatus] = (
            self.ts_update_status.all()
        )
        needs_update = {
            obj.section: obj.need_update()
            for obj in existing_sections
            if obj.section in sections_needs_update
        }
        sections_needs_update.update(needs_update)
        return _NeedsUpdate(section_map=sections_needs_update)

    def reset_update_status(self, section: UpdateSection) -> "OwnerUpdateStatus":
        """Reset the status of a given update section and return it."""
        update_status_obj: OwnerUpdateStatus = self.ts_update_status.get_or_create(
            section=section,
        )[0]
        update_status_obj.reset()
        return update_status_obj

    def reset_has_token_error(self) -> None:
        """Reset the has_token_error flag for this corporation."""
        self.ts_update_status.filter(
            has_token_error=True,
        ).update(
            has_token_error=False,
        )

    def update_section_if_changed(
        self,
        section: UpdateSection,
        fetch_func: Callable,
        force_refresh: bool = False,
    ):
        """Update the status of a specific section if it has changed."""
        section = self.UpdateSection(section)
        try:
            data = fetch_func(owner=self, force_refresh=force_refresh)
            logger.debug("%s: Update has changed, section: %s", self, section.label)
        except HTTPInternalServerError as exc:
            logger.debug("%s: Update has an HTTP internal server error: %s", self, exc)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except NotModifiedError:
            logger.debug("%s: Update has not changed, section: %s", self, section.label)
            return UpdateSectionResult(is_changed=False, is_updated=False)
        except HTTPGatewayTimeoutError:
            logger.debug(
                "%s: Update has a gateway timeout error, section: %s",
                self,
                section.label,
            )
            return UpdateSectionResult(is_changed=False, is_updated=False)
        return UpdateSectionResult(
            is_changed=True,
            is_updated=True,
            data=data,
        )

    def update_section_log(
        self,
        section: UpdateSection,
        is_success: bool,
        is_updated: bool = False,
        error_message: str = None,
    ) -> None:
        """Update the status of a specific section."""
        error_message = error_message if error_message else ""
        defaults = {
            "is_success": is_success,
            "error_message": error_message,
            "has_token_error": False,
            "last_run_finished_at": timezone.now(),
        }
        obj: OwnerUpdateStatus = self.ts_update_status.update_or_create(
            section=section,
            defaults=defaults,
        )[0]
        if is_updated:
            obj.last_update_at = obj.last_run_at
            obj.last_update_finished_at = timezone.now()
            obj.save()
        status = "successfully" if is_success else "with errors"
        logger.info("%s: %s Update run completed %s", self, section.label, status)

    def perform_update_status(
        self, section: UpdateSection, method: Callable, *args, **kwargs
    ) -> UpdateSectionResult:
        """Perform update status."""
        try:
            result = method(*args, **kwargs)
        except Exception as exc:
            # TODO ADD DISCORD NOTIFICATION?
            error_message = f"{type(exc).__name__}: {str(exc)}"
            is_token_error = isinstance(exc, (TokenError))
            logger.error(
                "%s: %s: Error during update status: %s",
                self,
                section.label,
                error_message,
                exc_info=not is_token_error,  # do not log token errors
            )
            self.ts_update_status.update_or_create(
                section=section,
                defaults={
                    "is_success": False,
                    "error_message": error_message,
                    "has_token_error": is_token_error,
                    "last_update_at": timezone.now(),
                },
            )
            raise exc
        return result

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax System Audit")
        verbose_name_plural = _("Tax System Audits")


class Members(models.Model):
    """Tax System Member model for app"""

    class States(models.TextChoices):
        ACTIVE = "active", _("Active")
        MISSING = "missing", _("Missing")
        NOACCOUNT = "noaccount", _("Unregistered")
        IS_ALT = "is_alt", _("Is Alt")

    character_name = models.CharField(max_length=100, db_index=True)

    character_id = models.PositiveIntegerField(primary_key=True)

    owner = models.ForeignKey(
        OwnerAudit, on_delete=models.CASCADE, related_name="ts_members"
    )

    status = models.CharField(
        _("Status"), max_length=10, choices=States.choices, blank=True, default="active"
    )

    logon = models.DateTimeField(null=True, blank=True)

    logged_off = models.DateTimeField(null=True, blank=True)

    joined = models.DateTimeField(null=True, blank=True)

    notice = models.TextField(null=True, blank=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Member System")
        verbose_name_plural = _("Tax Member Systems")

    def __str__(self):
        return f"{self.character_name} - {self.character_id}"

    objects = MembersManager()

    @property
    def is_active(self) -> bool:
        return self.status == self.States.ACTIVE

    @property
    def is_missing(self) -> bool:
        return self.status == self.States.MISSING

    @property
    def is_noaccount(self) -> bool:
        return self.status == self.States.NOACCOUNT

    @property
    def is_alt(self) -> bool:
        return self.status == self.States.IS_ALT

    @property
    def is_faulty(self) -> bool:
        return self.status in [self.States.MISSING, self.States.NOACCOUNT]


class PaymentSystem(models.Model):
    """Tax Payment System model for app"""

    class Status(models.TextChoices):
        ACTIVE = "active", _("Active")
        INACTIVE = "inactive", _("Inactive")
        DEACTIVATED = "deactivated", _("Deactivated")
        MISSING = "missing", _("Missing")

        def html(self, text=False) -> str:
            """Return the badge name for the status."""
            if text:
                return mark_safe(
                    f"<span class='badge bg-{self.color()}' data-tooltip-toggle='taxsystem-tooltip'>{self.label}</span>"
                )
            return mark_safe(
                f"<span class='badge bg-{self.color()}' data-tooltip-toggle='taxsystem-tooltip'>{self.icons()}</span>"
            )

        def color(self) -> str:
            """Return bootstrap corresponding icon class."""
            status_map = {
                self.ACTIVE: "success",
                self.INACTIVE: "warning",
                self.DEACTIVATED: "danger",
                self.MISSING: "info",
            }
            return status_map.get(self, "secondary")

        def icons(self) -> str:
            """Return description for an enum object."""
            status_map = {
                self.ACTIVE: "<i class='fas fa-check'></i>",
                self.INACTIVE: "<i class='fas fa-user-slash'></i>",
                self.DEACTIVATED: "<i class='fas fa-user-clock'></i>",
                self.MISSING: "<i class='fas fa-question'></i> ",
            }
            return status_map.get(self, "")

    class Paid(models.TextChoices):
        PAID = "paid", _("Paid")
        UNPAID = "unpaid", _("Unpaid")

    name = models.CharField(
        max_length=100,
    )

    owner = models.ForeignKey(
        OwnerAudit, on_delete=models.CASCADE, related_name="ts_payment_system"
    )

    user = models.OneToOneField(User, on_delete=models.CASCADE, related_name="+")

    date = models.DateTimeField(auto_now_add=True, null=True, blank=True)

    status = models.CharField(
        max_length=16,
        choices=Status.choices,
        blank=True,
        default=Status.ACTIVE,
    )

    deposit = models.DecimalField(
        max_digits=16,
        decimal_places=0,
        default=0,
        help_text=_("Deposit Pool in ISK. Max 16 Digits"),
        validators=[
            MaxValueValidator(9999999999999999),
            MinValueValidator(-9999999999999999),
        ],
    )

    last_paid = models.DateTimeField(null=True, blank=True)

    notice = models.TextField(null=True, blank=True)

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Payment System")
        verbose_name_plural = _("Tax Payment Systems")

    def __str__(self):
        return f"{self.name} - {self.date} - {self.status}"

    def get_payment_status(self) -> str:
        return self.get_status_display()

    def get_alt_ids(self) -> list[int]:
        return list(
            self.user.character_ownerships.all().values_list(
                "character__character_id", flat=True
            )
        )

    @property
    def is_active(self) -> bool:
        return self.status == self.Status.ACTIVE

    @property
    def is_inactive(self) -> bool:
        return self.status == self.Status.INACTIVE

    @property
    def is_deactivated(self) -> bool:
        return self.status == self.Status.DEACTIVATED

    @property
    def is_missing(self) -> bool:
        return self.status == self.Status.MISSING

    @property
    def has_paid(self) -> bool:
        """Return True if user has paid."""
        if self.deposit >= self.owner.tax_amount:
            return True
        if self.last_paid and self.deposit >= 0:
            return timezone.now() - self.last_paid < timezone.timedelta(
                days=self.owner.tax_period
            )
        return False

    @property
    def deposit_html(self) -> str:
        if self.deposit < 0:
            # Make text red for negative deposits
            return f"<span class='text-danger'>{intcomma(self.deposit, use_l10n=True)} ISK</span>"
        if self.deposit > 0:
            return f"<span class='text-success'>{intcomma(self.deposit, use_l10n=True)} ISK</span>"
        return (
            f"{intcomma(self.deposit, use_l10n=True)} ISK" if self.deposit else "0 ISK"
        )

    def has_paid_icon(self, badge=False, text=False) -> str:
        """Return the HTML icon for has_paid."""
        color = "success" if self.has_paid else "danger"

        if self.has_paid:
            html = f"<i class='fas fa-check' title='{self.Paid('paid').label}' data-tooltip-toggle='taxsystem-tooltip'></i>"
        else:
            html = f"<i class='fas fa-times' title='{self.Paid('unpaid').label}' data-tooltip-toggle='taxsystem-tooltip'></i>"

        if text:
            html += f" {self.Paid('paid').label if self.has_paid else self.Paid('unpaid').label}"

        if badge:
            html = mark_safe(f"<span class='badge bg-{color}'>{html}</span>")
        return {
            "display": html,
            "sort": _("Yes") if self.has_paid else _("No"),
            "raw": self.has_paid,
        }

    objects = PaymentSystemManager()


class Payments(models.Model):
    """Tax Payments model for app"""

    class RequestStatus(models.TextChoices):
        APPROVED = "approved", _("Approved")
        PENDING = "pending", _("Pending")
        REJECTED = "rejected", _("Rejected")
        NEEDS_APPROVAL = "needs_approval", _("Requires Auditor")

    name = models.CharField(max_length=100)

    entry_id = models.BigIntegerField()

    account = models.ForeignKey(
        PaymentSystem, on_delete=models.CASCADE, related_name="ts_payments"
    )

    corporation_id = models.IntegerField(null=True, blank=True)

    amount = models.DecimalField(max_digits=12, decimal_places=0)

    date = models.DateTimeField(null=True, blank=True)

    reason = models.TextField(null=True, blank=True)

    request_status = models.CharField(
        max_length=16,
        choices=RequestStatus.choices,
        default=RequestStatus.PENDING,
        verbose_name=_("Request Status"),
    )

    reviser = models.CharField(
        max_length=100,
        blank=True,
        default="",
        help_text=_("Reviser that approved or rejected the payment"),
    )

    class Meta:
        default_permissions = ()
        verbose_name = _("Tax Payments")
        verbose_name_plural = _("Tax Payments")

    @property
    def is_automatic(self) -> bool:
        return self.reviser == "System"

    @property
    def is_pending(self) -> bool:
        return self.request_status == self.RequestStatus.PENDING

    @property
    def is_needs_approval(self) -> bool:
        return self.request_status == self.RequestStatus.NEEDS_APPROVAL

    @property
    def is_approved(self) -> bool:
        return self.request_status == self.RequestStatus.APPROVED

    @property
    def is_rejected(self) -> bool:
        return self.request_status == self.RequestStatus.REJECTED

    @property
    def character_id(self) -> int:
        """Return the character ID of the user who made the payment or first OwnershipRecord."""
        try:
            character_id = self.account.user.profile.main_character.character_id
        except AttributeError:
            character = OwnershipRecord.objects.filter(user=self.account.user).first()
            character_id = character.character.character_id
        return character_id

    @property
    def division(self) -> "CorporationWalletJournalEntry":
        """Return the division name of the payment."""
        # pylint: disable=import-outside-toplevel
        # AA TaxSystem
        from taxsystem.models.wallet import CorporationWalletJournalEntry

        journal = CorporationWalletJournalEntry.objects.filter(
            entry_id=self.entry_id
        ).first()
        if not journal:
            return "N/A"
        return journal.division.name

    def __str__(self):
        return (
            f"{self.account.name} - {self.date} - {self.amount} - {self.request_status}"
        )

    def get_request_status(self) -> str:
        return self.get_request_status_display()

    @property
    def formatted_payment_date(self) -> str:
        if self.date:
            return timezone.localtime(self.date).strftime("%Y-%m-%d %H:%M:%S")
        return _("No date")

    objects = PaymentsManager()


class OwnerUpdateStatus(models.Model):
    """A Model to track the status of the last update."""

    owner = models.ForeignKey(
        OwnerAudit, on_delete=models.CASCADE, related_name="ts_update_status"
    )
    section = models.CharField(
        max_length=32, choices=OwnerAudit.UpdateSection.choices, db_index=True
    )
    is_success = models.BooleanField(default=None, null=True, db_index=True)
    error_message = models.TextField()
    has_token_error = models.BooleanField(default=False)

    last_run_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been started at this time",
    )
    last_run_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last run has been successful finished at this time",
    )
    last_update_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been started at this time",
    )
    last_update_finished_at = models.DateTimeField(
        default=None,
        null=True,
        db_index=True,
        help_text="Last update has been successful finished at this time",
    )

    class Meta:
        default_permissions = ()

    def __str__(self) -> str:
        return f"{self.owner} - {self.section} - {self.is_success}"

    def need_update(self) -> bool:
        """Check if the update is needed."""
        if not self.is_success or not self.last_update_finished_at:
            needs_update = True
        else:
            section_time_stale = app_settings.TAXSYSTEM_STALE_TYPES.get(
                self.section, 60
            )
            stale = timezone.now() - timezone.timedelta(minutes=section_time_stale)
            needs_update = self.last_run_finished_at <= stale

        if needs_update and self.has_token_error:
            logger.info(
                "%s: Ignoring update because of token error, section: %s",
                self.owner,
                self.section,
            )
            needs_update = False

        return needs_update

    def reset(self) -> None:
        """Reset this update status."""
        self.is_success = None
        self.error_message = ""
        self.has_token_error = False
        self.last_run_at = timezone.now()
        self.last_run_finished_at = None
        self.save()



# Standard Library
import logging

# Third Party
from ninja import NinjaAPI

# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.shortcuts import render
from django.utils.html import format_html
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.api.helpers import get_character_permissions, get_manage_corporation
from taxsystem.helpers import lazy
from taxsystem.models.logs import PaymentHistory
from taxsystem.models.tax import Payments, PaymentSystem

logger = logging.getLogger(__name__)


class CharacterApiEndpoints:
    tags = ["Character Tax System"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/character/{character_id}/payment/{pk}/view/details/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        # pylint: disable=too-many-locals
        def get_payment_details(
            request, corporation_id: int, character_id: int, pk: int
        ):
            owner, perms = get_manage_corporation(request, corporation_id)
            perms = perms or get_character_permissions(request, character_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 403, "Permission Denied"

            try:
                payment = Payments.objects.get(
                    pk=pk,
                    account__owner=owner,
                )
                account = PaymentSystem.objects.get(
                    user=payment.account.user,
                    owner=owner,
                )
            except Payments.DoesNotExist:
                return 404, "Payment Not Found"

            # Create a dict for the character
            paymentdetails = {
                "title": "Payments for",
                "character_id": character_id,
                "character_portrait": lazy.get_character_portrait_url(
                    character_id, size=32, as_html=True
                ),
                "character_name": payment.account.name,
                "payment_system": {},
                "payments": {},
            }

            # Create a dict for the payment system
            if account.is_active:
                status = lazy.generate_icon(
                    color="success", icon="fas fa-check", size=24, title=_("Active")
                )
            elif account.is_deactivated:
                status = lazy.generate_icon(
                    color="danger",
                    icon="fas fa-user-clock",
                    size=24,
                    title=_("Deactivated"),
                )
            else:
                status = lazy.generate_icon(
                    color="warning",
                    icon="fas fa-user-slash",
                    size=24,
                    title=_("Inactive"),
                )

            account_dict = {
                "account_id": account.pk,
                "corporation": account.owner.name,
                "name": account.name,
                "payment_pool": f"{intcomma(account.deposit)} ISK",
                "payment_status": PaymentSystem.Status(account.status).html(),
                "status": status,
            }

            payment_history_dict = {}

            payments_history = PaymentHistory.objects.filter(
                payment=payment,
            ).order_by("-date")

            for log in payments_history:
                log_dict = {
                    "log_id": log.pk,
                    "reviser": log.user,
                    "date": log.date,
                    "action": log.get_action_display(),
                    "comment": log.get_comment_display(),
                    "status": log.get_new_status_display(),
                }
                payment_history_dict[log.pk] = log_dict

            # Create Status Display
            payment_status = "<div class='text-center alert"
            if payment.is_pending:
                payment_status += " alert-warning'>"
            elif payment.is_needs_approval:
                payment_status += " alert-info'>"
            elif payment.is_approved:
                payment_status += " alert-success'>"
            elif payment.is_rejected:
                payment_status += " alert-danger'>"
            payment_status += f"{payment.get_request_status_display()}</div>"

            # Create a dict for each payment
            payment_dict = {
                "payment_id": payment.pk,
                "amount": f"{intcomma(payment.amount)} ISK",
                "date": payment.formatted_payment_date,
                "status": format_html(payment_status),
                "reviser": payment.reviser,
                "division": payment.division,
                "reason": payment.reason,
            }

            # Add payments to the character dict
            paymentdetails["payment"] = payment_dict
            paymentdetails["payment_system"] = account_dict
            paymentdetails["payment_history"] = payment_history_dict

            context = {
                "entity_pk": corporation_id,
                "entity_type": "character",
                "character": paymentdetails,
            }

            return render(
                request,
                "taxsystem/modals/view_payment_details.html",
                context,
            )



# Standard Library
import json
from pathlib import Path

# Alliance Auth (External Libs)
from eveuniverse.tools.testdata import load_testdata_from_dict


def _load_eveuniverse_from_file():
    with open(Path(__file__).parent / "eveuniverse.json", encoding="utf-8") as fp:
        return json.load(fp)


eveuniverse_testdata = _load_eveuniverse_from_file()


def load_eveuniverse():
    load_testdata_from_dict(eveuniverse_testdata)



"""Initialize the app"""

__version__ = "0.7.2"
__title__ = "Tax System"

__package_name__ = "aa-taxsystem"

__app_name_useragent__ = "AA-TaxSystem"
__github_url__ = f"https://github.com/Geuthur/{__package_name__}"



# Standard Library
from typing import TYPE_CHECKING

# Django
from django.db import models, transaction
from django.db.models import Case, Count, Q, Value, When

# Alliance Auth
from allianceauth.authentication.models import UserProfile
from allianceauth.services.hooks import get_extension_logger

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag
from eveuniverse.models import EveEntity

# AA TaxSystem
from taxsystem import __title__
from taxsystem.decorators import log_timing
from taxsystem.providers import esi
from taxsystem.task_helpers.etag_helpers import etag_results

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

if TYPE_CHECKING:
    # AA TaxSystem
    from taxsystem.models.tax import OwnerAudit


class OwnerAuditQuerySet(models.QuerySet):
    def visible_to(self, user):
        """Get all corps visible to the user."""
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.manage_corps"):
            logger.debug("Returning all corps for Tax Audit Manager %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            corp_ids = user.character_ownerships.all().values_list(
                "character__corporation_id", flat=True
            )
            queries = [models.Q(corporation__corporation_id__in=corp_ids)]

            logger.debug(
                "%s queries for user %s visible corporations.", len(queries), user
            )

            query = queries.pop()
            for q in queries:
                query |= q
            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def manage_to(self, user):
        """Get all corps that the user can manage."""
        # superusers get all visible
        if user.is_superuser:
            logger.debug(
                "Returning all corps for superuser %s.",
                user,
            )
            return self

        if user.has_perm("taxsystem.manage_corps"):
            logger.debug("Returning all corps for Tax Audit Manager %s.", user)
            return self

        try:
            char = user.profile.main_character
            assert char
            query = None

            if user.has_perm("taxsystem.manage_own_corp"):
                query = models.Q(corporation__corporation_id=char.corporation_id)

            logger.debug("Returning own corps for User %s.", user)

            if query is None:
                return self.none()

            return self.filter(query)
        except AssertionError:
            logger.debug("User %s has no main character. Nothing visible.", user)
            return self.none()

    def annotate_total_update_status_user(self, user):
        """Get the total update status for the given user."""
        char = user.profile.main_character
        assert char

        query = models.Q(character__character_ownership__user=user)

        return self.filter(query).annotate_total_update_status()

    def annotate_total_update_status(self):
        """Get the total update status."""
        # pylint: disable=import-outside-toplevel
        # AA TaxSystem
        from taxsystem.models.tax import OwnerAudit

        sections = OwnerAudit.UpdateSection.get_sections()
        num_sections_total = len(sections)
        qs = (
            self.annotate(
                num_sections_total=Count(
                    "ts_update_status",
                    filter=Q(ts_update_status__section__in=sections),
                )
            )
            .annotate(
                num_sections_ok=Count(
                    "ts_update_status",
                    filter=Q(
                        ts_update_status__section__in=sections,
                        ts_update_status__is_success=True,
                    ),
                )
            )
            .annotate(
                num_sections_failed=Count(
                    "ts_update_status",
                    filter=Q(
                        ts_update_status__section__in=sections,
                        ts_update_status__is_success=False,
                    ),
                )
            )
            .annotate(
                num_sections_token_error=Count(
                    "ts_update_status",
                    filter=Q(
                        ts_update_status__section__in=sections,
                        ts_update_status__has_token_error=True,
                    ),
                )
            )
            # pylint: disable=no-member
            .annotate(
                total_update_status=Case(
                    When(
                        active=False,
                        then=Value(OwnerAudit.UpdateStatus.DISABLED),
                    ),
                    When(
                        num_sections_token_error=1,
                        then=Value(OwnerAudit.UpdateStatus.TOKEN_ERROR),
                    ),
                    When(
                        num_sections_failed__gt=0,
                        then=Value(OwnerAudit.UpdateStatus.ERROR),
                    ),
                    When(
                        num_sections_ok=num_sections_total,
                        then=Value(OwnerAudit.UpdateStatus.OK),
                    ),
                    When(
                        num_sections_total__lt=num_sections_total,
                        then=Value(OwnerAudit.UpdateStatus.INCOMPLETE),
                    ),
                    default=Value(OwnerAudit.UpdateStatus.IN_PROGRESS),
                )
            )
        )

        return qs

    def disable_characters_with_no_owner(self) -> int:
        """Disable characters which have no owner. Return count of disabled characters."""
        orphaned_characters = self.filter(
            character__character_ownership__isnull=True, active=True
        )
        if orphaned_characters.exists():
            orphans = list(
                orphaned_characters.values_list(
                    "character__character_name", flat=True
                ).order_by("character__character_name")
            )
            orphaned_characters.update(active=False)
            logger.info(
                "Disabled %d characters which do not belong to a user: %s",
                len(orphans),
                ", ".join(orphans),
            )
            return len(orphans)
        return 0


class OwnerAuditManagerBase(models.Manager):
    def visible_to(self, user):
        return self.get_queryset().visible_to(user)

    def manage_to(self, user):
        return self.get_queryset().manage_to(user)


OwnerAuditManager = OwnerAuditManagerBase.from_queryset(OwnerAuditQuerySet)


class MembersQuerySet(models.QuerySet):
    pass


class MembersManagerBase(models.Manager):
    @log_timing(logger)
    def update_or_create_esi(
        self, owner: "OwnerAudit", force_refresh: bool = False
    ) -> None:
        """Update or Create a Members from ESI data."""
        return owner.update_section_if_changed(
            section=owner.UpdateSection.MEMBERS,
            fetch_func=self._fetch_esi_data,
            force_refresh=force_refresh,
        )

    def _fetch_esi_data(self, owner: "OwnerAudit", force_refresh: bool = False) -> None:
        """Fetch Members entries from ESI data."""
        req_scopes = [
            "esi-corporations.read_corporation_membership.v1",
            "esi-corporations.track_members.v1",
        ]
        req_roles = ["CEO", "Director"]

        token = owner.get_token(scopes=req_scopes, req_roles=req_roles)

        # Check Payment Accounts
        self._check_payment_accounts(owner)

        members_ob = (
            esi.client.Corporation.get_corporations_corporation_id_membertracking(
                corporation_id=owner.corporation.corporation_id,
            )
        )

        members_items = etag_results(members_ob, token, force_refresh=force_refresh)

        self._update_or_create_objs(
            owner,
            members_items,
        )

    @transaction.atomic()
    # pylint: disable=too-many-locals
    def _update_or_create_objs(
        self,
        owner: "OwnerAudit",
        objs: list,
    ) -> None:
        """Update or Create Members entries from objs data."""
        _current_members_ids = set(
            self.filter(owner=owner).values_list("character_id", flat=True)
        )
        _esi_members_ids = [member.get("character_id") for member in objs]
        _old_members = []
        _new_members = []

        characters = EveEntity.objects.bulk_resolve_names(_esi_members_ids)

        for member in objs:
            character_id = member.get("character_id")
            joined = member.get("start_date")
            logon_date = member.get("logon_date")
            logged_off = member.get("logoff_date")
            character_name = characters.to_name(character_id)
            member_item = self.model(
                owner=owner,
                character_id=character_id,
                character_name=character_name,
                joined=joined,
                logon=logon_date,
                logged_off=logged_off,
                status=self.model.States.ACTIVE,
            )
            if character_id in _current_members_ids:
                _old_members.append(member_item)
            else:
                _new_members.append(member_item)

        # Set missing members
        old_member_ids = {member.character_id for member in _old_members}
        missing_members_ids = _current_members_ids - old_member_ids

        if missing_members_ids:
            self.filter(owner=owner, character_id__in=missing_members_ids).update(
                status=self.model.States.MISSING
            )
            logger.debug(
                "Marked %s missing members for: %s",
                len(missing_members_ids),
                owner.corporation.corporation_name,
            )
        if _old_members:
            self.bulk_update(
                _old_members,
                ["character_name", "status", "logon", "logged_off"],
            )
            logger.debug(
                "Updated %s members for: %s",
                len(_old_members),
                owner.corporation.corporation_name,
            )
        if _new_members:
            self.bulk_create(_new_members, ignore_conflicts=True)
            logger.debug(
                "Added %s new members for: %s",
                len(_new_members),
                owner.corporation.corporation_name,
            )

        # Update payment accounts
        self._update_payment_accounts(owner, _esi_members_ids)

        logger.info(
            "Corp %s - Old Members: %s, New Members: %s, Missing: %s",
            owner.name,
            len(_old_members),
            len(_new_members),
            len(missing_members_ids),
        )

    def _update_payment_accounts(self, owner: "OwnerAudit", members_ids: list[int]):
        """Update payment accounts for a corporation."""
        # pylint: disable=import-outside-toplevel
        # AA TaxSystem
        from taxsystem.models.tax import PaymentSystem

        logger.debug(
            "Updating Payment Accounts for: %s",
            owner.corporation.corporation_name,
        )

        accounts = UserProfile.objects.filter(
            main_character__isnull=False,
            main_character__corporation_id=owner.corporation.corporation_id,
        ).select_related(
            "user__profile__main_character",
            "main_character__character_ownership",
            "main_character__character_ownership__user__profile",
            "main_character__character_ownership__user__profile__main_character",
        )

        members = self.filter(owner=owner)

        if not accounts:
            logger.debug(
                "No valid accounts for: %s", owner.corporation.corporation_name
            )
            return "No Accounts"

        items = []

        for account in accounts:
            alts = set(
                account.user.character_ownerships.all().values_list(
                    "character__character_id", flat=True
                )
            )
            main = account.main_character

            # Change the status of members if they are alts
            relevant_alts = alts.intersection(members_ids)
            for alt in relevant_alts:
                members_ids.remove(alt)
                if alt != main.character_id:
                    # Update the status of the member to alt
                    members.filter(character_id=alt).update(
                        status=self.model.States.IS_ALT
                    )

            # Create or update a Payment System for the main character
            try:
                existing_payment_system = PaymentSystem.objects.get(
                    user=account.user, owner=owner
                )

                if existing_payment_system.status != PaymentSystem.Status.DEACTIVATED:
                    existing_payment_system.status = PaymentSystem.Status.ACTIVE
                    existing_payment_system.save()
            except PaymentSystem.DoesNotExist:
                items.append(
                    PaymentSystem(
                        name=main.character_name,
                        owner=owner,
                        user=account.user,
                        status=PaymentSystem.Status.ACTIVE,
                    )
                )

        if members_ids:
            # Mark members without accounts
            for member_id in members_ids:
                members.filter(character_id=member_id).update(
                    status=self.model.States.NOACCOUNT
                )

            logger.debug(
                "Marked %s members without accounts for: %s",
                len(members_ids),
                owner.corporation.corporation_name,
            )

        if items:
            PaymentSystem.objects.bulk_create(items, ignore_conflicts=True)
            logger.info(
                "Added %s new payment users for: %s",
                len(items),
                owner.corporation.corporation_name,
            )
        else:
            logger.debug(
                "No new payment user for: %s",
                owner.corporation.corporation_name,
            )

        return (
            "Finished payment system for %s",
            owner.corporation.corporation_name,
        )

    def _check_payment_accounts(self, owner: "OwnerAudit"):
        """Check payment accounts for a corporation."""
        # pylint: disable=import-outside-toplevel
        # AA TaxSystem
        from taxsystem.models.tax import OwnerAudit, PaymentSystem

        logger.debug(
            "Checking Payment Accounts for: %s",
            owner.corporation.corporation_name,
        )

        accounts = UserProfile.objects.filter(
            main_character__isnull=False,
        ).select_related(
            "user__profile__main_character",
            "main_character__character_ownership",
            "main_character__character_ownership__user__profile",
            "main_character__character_ownership__user__profile__main_character",
        )

        if not accounts:
            logger.debug(
                "No valid accounts for skipping Check: %s",
                owner.corporation.corporation_name,
            )
            return "No Accounts"

        for account in accounts:
            main_corporation_id = account.main_character.corporation_id

            try:
                payment_system = PaymentSystem.objects.get(
                    user=account.user, owner=owner
                )
                payment_system_corp_id = payment_system.owner.corporation.corporation_id
                # Check if the user is no longer in the same corporation
                if (
                    not payment_system.is_missing
                    and not payment_system_corp_id == main_corporation_id
                ):
                    payment_system.status = PaymentSystem.Status.MISSING
                    payment_system.save()
                    logger.info(
                        "User %s is no longer in Corp marked as Missing",
                        payment_system.name,
                    )
                # Check if the user changed to a existing corporation Payment System
                elif (
                    payment_system.is_missing
                    and payment_system_corp_id != main_corporation_id
                ):
                    try:
                        new_owner = OwnerAudit.objects.get(
                            corporation__corporation_id=main_corporation_id
                        )
                        payment_system.owner = new_owner
                        payment_system.deposit = 0
                        payment_system.status = PaymentSystem.Status.ACTIVE
                        payment_system.last_paid = None
                        payment_system.save()
                        logger.info(
                            "User %s is now in Corporation %s",
                            payment_system.name,
                            new_owner.corporation.corporation_name,
                        )
                    except owner.DoesNotExist:
                        continue
                elif (
                    payment_system.is_missing
                    and payment_system_corp_id == main_corporation_id
                ):
                    payment_system.status = PaymentSystem.Status.ACTIVE
                    payment_system.notice = None
                    payment_system.deposit = 0
                    payment_system.last_paid = None
                    payment_system.save()
                    logger.info(
                        "User %s is back in Corporation %s",
                        payment_system.name,
                        payment_system.owner.corporation.corporation_name,
                    )
            except PaymentSystem.DoesNotExist:
                logger.debug(
                    "No Payment System for %s - %s",
                    account.user.username,
                    owner.corporation.corporation_name,
                )
                continue
        return (
            "Finished checking Payment Accounts for %s",
            owner.corporation.corporation_name,
        )


MembersManager = MembersManagerBase.from_queryset(MembersQuerySet)



# Standard Library
from unittest.mock import Mock, patch

# Django
from django.core.cache import cache

# Alliance Auth (External Libs)
from app_utils.testing import NoSocketsTestCase

# AA TaxSystem
from taxsystem.task_helpers.etag_helpers import (
    MAX_ETAG_LIFE,
    HTTPNotModified,
    NotModifiedError,
    del_etag_header,
    etag_results,
    get_etag_header,
    get_etag_key,
    handle_etag_headers,
    handle_page_results,
    inject_etag_header,
    rem_etag_header,
    set_etag_header,
    stringify_params,
)

MODULE_PATH = "taxsystem.task_helpers.etag_helpers"


def clear_cache(cache_key):
    """Clear cache."""
    cache.delete(cache_key)


class TestEtagHelpers(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        clear_cache("taxsystem-cache_key")
        cls.headers = Mock()
        cls.headers.headers = {
            "ETag": "ETAG",
            "X-Pages": "1",
        }
        cls.operation = Mock()
        cls.operation._cache_key.return_value = "cache_key"
        cls.operation.operation.operation_id = "operation_id"
        cls.operation.future.request.headers = {
            "If-None-Match": "ABCD",
            "If-Modified-Since": "2023-10-01T00:00:00Z",
        }
        cls.operation.future.request.params = {
            "param1": "value1",
            "param2": "value2",
        }

        cls.headers_no_etag = Mock()
        cls.headers_no_etag.headers = {}
        cache.set("taxsystem-cache_key", "ABCD", MAX_ETAG_LIFE)

    def test_get_etag_key(self):
        result = get_etag_key(self.operation)
        self.assertEqual(result, "taxsystem-cache_key")

    def test_get_etag_header(self):
        # given
        cache.set("taxsystem-cache_key", "ABCD", MAX_ETAG_LIFE)
        # when
        result = get_etag_header(self.operation)
        # then
        self.assertEqual(result, "ABCD")

    def test_inject_etag_header(self):
        inject_etag_header(self.operation)
        self.assertEqual(self.operation.future.request.headers["If-None-Match"], "ABCD")

    def test_set_etag_header(self):
        result = set_etag_header(self.operation, self.headers)
        self.assertTrue(result)

        result = set_etag_header(self.operation, self.headers_no_etag)
        self.assertFalse(result)

    def test_stringify_params(self):
        result = stringify_params(self.operation)
        self.assertEqual(result, "param1: value1, param2: value2")


class TestEtagPageHandler(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        clear_cache("taxsystem-cache_key_2")
        cls.headers = Mock()
        cls.headers.headers = {
            "ETag": "ETAG",
            "X-Pages": "2",
        }
        cls.operation = Mock()
        cls.operation._cache_key.return_value = "cache_key_2"
        cls.operation.operation.operation_id = "operation_id"
        cls.operation.future.request.headers = {
            "If-None-Match": "DCBA",
            "If-Modified-Since": "2023-10-01T00:00:00Z",
        }
        cls.operation.future.request.params = {
            "param1": "value1",
            "param2": "value2",
        }
        cls.operation.result.return_value = ([], cls.headers)

        cls.headers_no_etag = Mock()
        cls.headers_no_etag.headers = {}
        cache.set("taxsystem-cache_key_2", "DCBA", MAX_ETAG_LIFE)

    def test_handle_page_results_should_be_success_with_page_3(self):
        results, current_page, total_pages = handle_page_results(
            self.operation, 1, 2, False, False
        )
        self.assertEqual(results, [])
        self.assertEqual(current_page, 3)
        self.assertEqual(total_pages, 2)

    @patch(MODULE_PATH + ".handle_etag_headers")
    def test_handle_page_results_should_raise_notmodifiederror(
        self, mock_handle_etag_headers
    ):
        # given
        mock_handle_etag_headers.side_effect = NotModifiedError()
        # when
        with self.assertRaises(NotModifiedError):
            handle_page_results(self.operation, 1, 2, False, False)

    @patch(MODULE_PATH + ".handle_etag_headers")
    def test_handle_page_results_should_raise_http_not_modified(
        self, mock_handle_etag_headers
    ):
        # given
        mock_response = Mock()
        mock_response.status_code = 304
        mock_response.headers = {"ETag": "ETAG", "X-Pages": "2"}
        mock_handle_etag_headers.side_effect = HTTPNotModified(mock_response)
        # when
        with self.assertRaises(
            NotModifiedError
        ):  # Raise NotModifiedError after raising HTTPNotModified
            handle_page_results(self.operation, 1, 2, False, False)



"""App Tasks"""

# Standard Library
import inspect
from collections.abc import Callable

# Third Party
from celery import chain, shared_task

# Alliance Auth
from allianceauth.services.hooks import get_extension_logger
from allianceauth.services.tasks import QueueOnce

# Alliance Auth (External Libs)
from app_utils.logging import LoggerAddTag

# AA TaxSystem
from taxsystem import __title__, app_settings
from taxsystem.decorators import when_esi_is_available
from taxsystem.models.tax import OwnerAudit

logger = LoggerAddTag(get_extension_logger(__name__), __title__)

MAX_RETRIES_DEFAULT = 3

# Default params for all tasks.
TASK_DEFAULTS = {
    "time_limit": app_settings.TAXSYSTEM_TASKS_TIME_LIMIT,
    "max_retries": MAX_RETRIES_DEFAULT,
}

# Default params for tasks that need run once only.
TASK_DEFAULTS_ONCE = {**TASK_DEFAULTS, **{"base": QueueOnce}}

_update_taxsystem_params = {
    **TASK_DEFAULTS_ONCE,
    **{"once": {"keys": ["owner_pk", "force_refresh"], "graceful": True}},
}


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_all_taxsytem(runs: int = 0, force_refresh: bool = False):
    """Update all taxsystem data"""
    corporations = OwnerAudit.objects.select_related("corporation").filter(active=1)
    for corporation in corporations:
        update_corporation.apply_async(
            args=[corporation.pk], kwargs={"force_refresh": force_refresh}
        )
        runs = runs + 1
    logger.info("Queued %s OwnerAudit Tasks", runs)


@shared_task(**TASK_DEFAULTS_ONCE)
@when_esi_is_available
def update_corporation(owner_pk, force_refresh=False):
    """Update a corporation"""
    owner = OwnerAudit.objects.prefetch_related("ts_update_status").get(pk=owner_pk)

    que = []
    priority = 7

    logger.debug(
        "Processing Audit Updates for %s",
        format(owner.corporation.corporation_name),
    )

    if force_refresh:
        # Reset Token Error if we are forcing a refresh
        owner.reset_has_token_error()

    needs_update = owner.calc_update_needed()

    if not needs_update and not force_refresh:
        logger.info("No updates needed for %s", owner.corporation.corporation_name)
        return

    sections = owner.UpdateSection.get_sections()

    for section in sections:
        # Skip sections that are not in the needs_update list
        if not force_refresh and not needs_update.for_section(section):
            logger.debug(
                "No updates needed for %s (%s)",
                owner.corporation.corporation_name,
                section,
            )
            continue

        task_name = f"update_owner_{section}"
        task = globals().get(task_name)
        que.append(
            task.si(owner.pk, force_refresh=force_refresh).set(priority=priority)
        )

    chain(que).apply_async()
    logger.debug(
        "Queued %s Audit Updates for %s",
        len(que),
        owner.corporation.corporation_name,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_division_names(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.DIVISION_NAMES,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_division(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.DIVISION,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_wallet(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.WALLET,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_members(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.MEMBERS,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_payments(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.PAYMENTS,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_payment_system(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.PAYMENT_SYSTEM,
        force_refresh=force_refresh,
    )


@shared_task(**_update_taxsystem_params)
def update_owner_payday(owner_pk: int, force_refresh: bool):
    return _update_owner_section(
        owner_pk,
        section=OwnerAudit.UpdateSection.PAYDAY,
        force_refresh=force_refresh,
    )


def _update_owner_section(owner_pk: int, section: str, force_refresh: bool):
    """Update a specific section of the character audit."""
    section = OwnerAudit.UpdateSection(section)
    corporation = OwnerAudit.objects.get(pk=owner_pk)
    logger.debug(
        "Updating %s for %s", section.label, corporation.corporation.corporation_name
    )

    corporation.reset_update_status(section)

    method: Callable = getattr(corporation, section.method_name)
    method_signature = inspect.signature(method)

    if "force_refresh" in method_signature.parameters:
        kwargs = {"force_refresh": force_refresh}
    else:
        kwargs = {}

    result = corporation.perform_update_status(section, method, **kwargs)
    corporation.update_section_log(
        section, is_success=True, is_updated=result.is_updated
    )


@shared_task(**TASK_DEFAULTS_ONCE)
def clear_all_etags():
    logger.debug("Clearing all etags")
    try:
        # Third Party
        # pylint: disable=import-outside-toplevel
        from django_redis import get_redis_connection

        _client = get_redis_connection("default")
    except (NotImplementedError, ModuleNotFoundError):
        # Django
        # pylint: disable=import-outside-toplevel
        from django.core.cache import caches

        default_cache = caches["default"]
        _client = default_cache.get_master_client()
    keys = _client.keys(":?:taxsystem-*")
    if keys:
        deleted = _client.delete(*keys)
        logger.info("Deleted %s etag keys", deleted)
    else:
        logger.info("No etag keys to delete")



# Django
from django.contrib.humanize.templatetags.humanize import intcomma
from django.urls import reverse
from django.utils.html import format_html
from django.utils.translation import gettext as _

# AA TaxSystem
from taxsystem.api.helpers import generate_button, generate_settings
from taxsystem.models.tax import PaymentSystem

PAYMENT_SYSTEM_TEMPLATE = "taxsystem/partials/form/button.html"


def _payments_info(corporation_id, user: PaymentSystem, perms, request):
    if not perms:
        return ""

    view_payments = generate_button(
        corporation_id,
        "taxsystem/partials/form/button.html",
        user,
        {
            "title": _("View Payments"),
            "icon": "fas fa-info",
            "color": "primary",
            "text": _("View Payments"),
            "modal": "modalViewPaymentsContainer",
            "action": f"/taxsystem/api/corporation/{corporation_id}/character/{user.user.profile.main_character.character_id}/view/payments/",
            "ajax": "ajax_payments",
        },
        request,
    )

    return format_html(f"{intcomma(user.deposit, use_l10n=True)} ISK {view_payments}")


def _switch_user(
    actions: list, corporation_id: int, payment_system: PaymentSystem, request
):
    url = reverse(
        viewname="taxsystem:switch_user",
        kwargs={
            "corporation_id": corporation_id,
            "payment_system_pk": payment_system.pk,
        },
    )

    if payment_system.is_active:
        confirm_text = (
            _("Are you sure to Confirm")
            + f"?<br><span class='fw-bold'>{payment_system.name} "
            + _("Deactivate")
            + "</span>"
        )
        title = _("Deactivate User")
        icon = "fas fa-eye-low-vision"
        color = "warning"
    else:
        confirm_text = (
            _("Are you sure to Confirm")
            + f"?<br><span class='fw-bold'>{payment_system.name} "
            + _("Activate")
            + "</span>"
        )
        title = _("Activate User")
        icon = "fas fa-eye"
        color = "success"

    settings = generate_settings(
        title=title,
        icon=icon,
        color=color,
        text=confirm_text,
        modal="paymentsystem-switchuser",
        action=url,
        ajax="action",
    )
    actions.append(
        generate_button(
            corporation_id, PAYMENT_SYSTEM_TEMPLATE, payment_system, settings, request
        )
    )

    return actions


def _add_payment(
    actions: list, corporation_id: int, payment_system: PaymentSystem, request
):
    actions.append(
        generate_button(
            corporation_id=corporation_id,
            template="taxsystem/partials/form/button.html",
            queryset=payment_system,
            settings={
                "title": _("Add Payment"),
                "icon": "fas fa-fas fa-dollar-sign",
                "color": "success",
                "text": _("Add Payment")
                + _("for")
                + f" {payment_system.user.username}",
                "modal": "payments-add",
                "action": reverse(
                    viewname="taxsystem:add_payment",
                    kwargs={
                        "corporation_id": corporation_id,
                        "payment_system_pk": payment_system.pk,
                    },
                ),
                "ajax": "action",
            },
            request=request,
        )
    )
    return actions


def payment_system_actions(
    corporation_id, payment_system: PaymentSystem, perms, request
):
    # Check if user has permission to view the actions
    if not perms:
        return ""

    # Generate the buttons
    actions = []

    # Add Switch User Button
    _add_payment(
        actions=actions,
        corporation_id=corporation_id,
        payment_system=payment_system,
        request=request,
    )
    _switch_user(
        actions=actions,
        corporation_id=corporation_id,
        payment_system=payment_system,
        request=request,
    )

    actions_html = format_html("".join(actions))
    return format_html('<div class="d-flex justify-content-end">{}</div>', actions_html)



# Standard Library
import logging

# Third Party
from ninja import NinjaAPI

# Django
from django.utils.translation import gettext_lazy as _

# AA TaxSystem
from taxsystem.api.helpers import get_corporation, get_manage_corporation
from taxsystem.api.taxsystem.helpers.own_payments import _own_payments_actions
from taxsystem.api.taxsystem.helpers.payments import payments_actions
from taxsystem.helpers import lazy
from taxsystem.models.tax import Payments, PaymentSystem

logger = logging.getLogger(__name__)


class CorporationApiEndpoints:
    tags = ["Corporation Tax System"]

    # pylint: disable=too-many-statements
    def __init__(self, api: NinjaAPI):
        @api.get(
            "corporation/{corporation_id}/view/payments/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_payments(request, corporation_id: int):
            owner, perms = get_manage_corporation(request, corporation_id)

            if owner is None:
                return 404, "Corporation Not Found"

            if perms is False:
                return 404, "Permission Denied"

            payments = (
                Payments.objects.filter(
                    account__owner=owner,
                    corporation_id=owner.corporation.corporation_id,
                )
                .select_related("account")
                .order_by("-date")
            )

            payments_dict = {}

            for payment in payments:
                try:
                    character_portrait = lazy.get_character_portrait_url(
                        payment.character_id, size=32, as_html=True
                    )
                except ValueError:
                    character_portrait = ""

                actions = payments_actions(corporation_id, payment, perms, request)

                payments_dict[payment.pk] = {
                    "payment_id": payment.pk,
                    "date": payment.formatted_payment_date,
                    "character_portrait": character_portrait,
                    "character_name": payment.account.name,
                    "amount": payment.amount,
                    "request_status": payment.get_request_status_display(),
                    "division": payment.division,
                    "reason": payment.reason,
                    "actions": actions,
                }

            output = []
            output.append({"corporation": payments_dict})

            return output

        @api.get(
            "corporation/{corporation_id}/view/own-payments/",
            response={200: list, 403: str, 404: str},
            tags=self.tags,
        )
        def get_own_payments(request, corporation_id: int):
            corp = get_corporation(request, corporation_id)

            if corp is None:
                return 404, "Corporation Not Found"

            account = PaymentSystem.objects.get(owner=corp, user=request.user)

            payments = (
                Payments.objects.filter(
                    account__owner=corp,
                    account=account,
                    corporation_id=corp.corporation.corporation_id,
                )
                .select_related("account")
                .order_by("-date")
            )

            own_payments_dict = {}

            for payment in payments:
                actions = _own_payments_actions(corporation_id, payment, request)

                own_payments_dict[payment.pk] = {
                    "payment_id": payment.pk,
                    "date": payment.formatted_payment_date,
                    "character_name": payment.account.name,
                    "amount": payment.amount,
                    "request_status": payment.get_request_status_display(),
                    "division": payment.division,
                    "reason": payment.reason,
                    "actions": actions,
                }

            output = []
            output.append({"corporation": own_payments_dict})

            return output
